#!/usr/bin/python3

from gi.repository import Gtk, GObject, GLib

from util import trackers

class _fixedViewport(Gtk.Viewport):
    """
    This is needed by MarqueeLabel to restrict the size of
    our label, and cause the viewport to actually be functional.
    Otherwise, the text is trimmed, but no scrolling occurs.
    """
    def __init__(self):
        super(_fixedViewport, self).__init__()

        self.set_shadow_type(Gtk.ShadowType.NONE)

    def do_get_preferred_width(self):
        return 400, 400

class MarqueeLabel(Gtk.Stack):
    """
    A scrolling label for the PlayerControl - it uses the defined
    pattern as a mapping between elapsed time and the hadjustment
    of the _fixedViewport.  If the label text is wider than the
    widget's actual width, we will scroll according to this map
    over the course of 15 seconds.

    It roughly translates to:
         0.0 -  2.0 seconds: no movement.
         2.0 - 10.0 seconds: gradually scroll to max adjustment of the viewport.
        10.0 - 12.0 seconds: no movement.
        12.0 - 15.0 seconds: scroll back to the starting position.

    This widget also implements a stack (similar to the one in MonitorView) which
    provides for a smooth label crossfade when track info changes.
    """
    PATTERN = [( 0.0, 0.0),
               ( 2.0, 0.0),
               (10.0, 1.0),
               (12.0, 1.0),
               (15.0, 0.0)]

    LENGTH = len(PATTERN)

    def __init__(self, text):
        super(MarqueeLabel, self).__init__()

        self.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
        self.set_transition_duration(300)

        self.tick_id = 0

        self.current = self._make_label(text)

        self.add(self.current)
        self.set_visible_child(self.current)

    def _make_label(self, text):
        vp = _fixedViewport()

        label = Gtk.Label(text)
        label.set_halign(Gtk.Align.START)

        vp.add(label)
        vp.show_all()

        return vp

    def set_text(self, text):
        if self.current.get_child().get_text() == text:
            return

        self.cancel_tick()

        self.queued = self._make_label(text)

        self.add(self.queued)
        self.set_visible_child(self.queued)

        tmp = self.current
        self.current = self.queued
        self.queued = None

        GObject.idle_add(tmp.destroy)

        if not self.current.get_realized():
            trackers.con_tracker_get().connect(self.current,
                                               "realize",
                                               self.on_current_realized)
        else:
            GObject.idle_add(self._marquee_idle)

    def on_current_realized(self, widget, data=None):
        GObject.idle_add(self._marquee_idle)

        trackers.con_tracker_get().disconnect(widget,
                                              "realize",
                                              self.on_current_realized)

    def cancel_tick(self):
        if self.tick_id > 0:
            self.remove_tick_callback(self.tick_id)
            self.tick_id = 0

    def _marquee_idle(self):
        self.hadjust = self.current.get_hadjustment()

        if (self.hadjust.get_upper() == self.hadjust.get_page_size()) == self.get_allocated_width():
            return False

        self.start_time = self.get_frame_clock().get_frame_time()
        self.end_time = self.start_time + (self.PATTERN[self.LENGTH - 1][0] * 1000 * 1000) # sec to ms to μs

        if self.tick_id == 0:
            self.tick_id = self.add_tick_callback(self._on_marquee_tick)

        self._marquee_step(self.start_time)

        return GLib.SOURCE_REMOVE

    def _on_marquee_tick(self, widget, clock, data=None):
        now = clock.get_frame_time()

        self._marquee_step(now)

        if now >= self.end_time:
            self.start_time = self.end_time
            self.end_time += (self.PATTERN[self.LENGTH - 1][0] * 1000 * 1000) # sec to ms to μs

        return GLib.SOURCE_CONTINUE

    def interpolate_point(self, now):
        point = ((now - self.start_time) / 1000 / 1000)

        i = 0
        while i < self.LENGTH:
            cindex, cval = self.PATTERN[i]

            if point > cindex:
                i += 1
                continue

            if point == cindex:
                return cval

            pindex, pval = self.PATTERN[i - 1]
            diff = cval - pval
            duration = cindex - pindex

            ratio = diff / duration
            additive = (point - pindex) * ratio
            return pval + additive

    def _marquee_step(self, now):
        if now < self.end_time:
            t = self.interpolate_point(now)
        else:
            t = self.PATTERN[self.LENGTH - 1][1]

        new_position = ((self.hadjust.get_upper() - self.hadjust.get_page_size()) * t)

        self.hadjust.set_value(new_position)
        self.queue_draw()



#!/usr/bin/python3

from gi.repository import Gtk, Gio, GLib, GObject
import re
import cairo
import signal

import status
from baseWindow import BaseWindow
from util import settings, utils, trackers

class WallpaperStack(Gtk.Stack):
    """
    WallpaperStack implements a crossfade when changing backgrounds.

    An initial image is made and added to the GtkStack.  When a new
    image is requested, it is created and added to the stack, then
    a crossfade transition is made to the new child.  The former
    visible stack child is then destroyed.  And this repeats.
    """
    def __init__(self):
        super(WallpaperStack, self).__init__()

        self.set_transition_type(Gtk.StackTransitionType.NONE)
        self.set_transition_duration(1000)

        self.initialized = False
        self.current = None
        self.queued = None

    def transition_to_image(self, image):
        """
        Queues a new image in the stack, and begins the transition to it.
        """
        self.queued = image
        self.queued.set_visible(True)

        trackers.con_tracker_get().connect_after(image,
                                                 "draw",
                                                 self.shade_wallpaper)

        self.add(self.queued)

        if not self.initialized:
            self.visible_image_changed()
            self.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
            self.initialized = True
            return

        self.set_visible_child(self.queued)
        GObject.timeout_add(2000, self.visible_image_changed)

    def visible_image_changed(self, data=None):
        if self.current is not None:
            tmp = self.current

            self.remove(tmp)
            tmp.destroy()

        self.current = self.queued
        self.queued = None

        return False

    def shade_wallpaper(self, widget, cr):
        """
        This draw callback adds a shade mask over the current
        image.  It is uniform when not Awake, and acquires a
        significant gradient vertically framing the unlock dialog
        when Awake.
        """
        cr.set_source_rgba(0.0, 0.0, 0.0, 0.7)
        cr.paint()
        return False

class MonitorView(BaseWindow):
    """
    A monitor-sized child of the stage that is responsible for displaying
    the currently-selected wallpaper or appropriate plug-in.
    """
    def __init__(self, index):
        super(MonitorView, self).__init__()

        self.monitor_index = index

        self.update_geometry()

        self.wallpaper_stack = WallpaperStack()
        self.wallpaper_stack.show()
        self.wallpaper_stack.set_halign(Gtk.Align.FILL)
        self.wallpaper_stack.set_valign(Gtk.Align.FILL)
        self.add(self.wallpaper_stack)

        self.show_all()

    def set_next_wallpaper_image(self, image):
        self.wallpaper_stack.transition_to_image(image)



# Generated file - DO NOT EDIT.  Edit config.py.in instead.

prefix="/usr"
datadir="/usr/share"
localedir=datadir+"/locale"
pkgdatadir="/usr/share/cinnamon-screensaver"
libdir="/usr/lib"
pkglibdir="/usr/lib/cinnamon-screensaver"
PACKAGE="cinnamon-screensaver"
VERSION="6.6.1"
GETTEXT_PACKAGE="cinnamon-screensaver"



#!/usr/bin/python3

from gi.repository import Gtk, GObject, Gdk

from util import trackers
import status

class BaseWindow(Gtk.Bin):
    """
    BaseWindow is the base class for all of the Stage GtkOverlay's immediate
    children.
    """

    def __init__(self, *args):
        super(BaseWindow, self).__init__()

        self.disabled = False

        c = Gdk.RGBA(0, 0, 0, 0)
        self.override_background_color (Gtk.StateFlags.NORMAL, c)

    def destroy_window(self):
        self.destroy()

    def update_geometry(self):
        if status.Spanned:
            self.rect = status.screen.get_screen_geometry()
        else:
            self.rect = status.screen.get_monitor_geometry(self.monitor_index)



#!/usr/bin/python3
# coding: utf-8

import gi

gi.require_version('CinnamonDesktop', '3.0')
from gi.repository import Gtk, Gdk, GdkPixbuf, GLib
import cairo

from util import trackers, settings
import singletons

class PasswordEntry(Gtk.Entry):
    """
    The GtkEntry where the user types their password.  It also
    implements a clickable imagine showing the currently chosen
    keyboard layout, and allowing switching of the layout.
    """
    def __init__(self):
        super(PasswordEntry, self).__init__(max_length=200)
        self.get_style_context().add_class("passwordentry")

        placeholder_text = _("Please enter your password...")
        self.set_width_chars(len(placeholder_text) + 1) # account for the flag
        self.set_has_frame(True)
        self.set_input_purpose(Gtk.InputPurpose.PASSWORD)
        self.set_visibility(False)
        self.set_property("caps-lock-warning", False)
        self.set_placeholder_text (placeholder_text)
        self.set_can_default(True)

        self.set_icon_from_icon_name(Gtk.EntryIconPosition.SECONDARY, "cinnamon-screensaver-view-reveal")
        trackers.con_tracker_get().connect(self, "icon-press", self.on_icon_pressed)

        self.placeholder_text = placeholder_text
        self.lockscreen_layout_source = None
        self.system_layout_source = None

        self.cinnamon = singletons.CinnamonClient
        self.cinnamon.connect("current-input-source-changed", self.on_current_layout_changed)
        self.cinnamon.connect("input-sources-changed", self.on_layout_sources_changed)
        self.on_layout_sources_changed(self.cinnamon)

        trackers.con_tracker_get().connect(self,
                                           "destroy",
                                           self.on_destroy)

    def on_draw(self, widget, cr, data=None):
        """
        GtkEntry always makes its icons menu-sized, no matter how much actual
        space is available for the image.  So, we use a transparent icon in
        update_layout_icon(), just so GtkEntry thinks there's an icon there,
        that way it allocates space for it, and responds to clicks in the area.
        """
        icon_rect = widget.get_icon_area(Gtk.EntryIconPosition.PRIMARY)
        x = icon_rect.x
        y = icon_rect.y + 2
        width = (icon_rect.width // 2) * 2
        height = icon_rect.height - 4

        handled = False
        if settings.get_show_flags():
            ui_scale = self.get_scale_factor()

            if self.lockscreen_layout_source.flag_name != "":
                filename = "/usr/share/iso-flag-png/%s.png" % self.lockscreen_layout_source.flag_name
                try:
                    pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(filename, -1, height * ui_scale)

                    logical_width = pixbuf.get_width() / ui_scale
                    logical_height = pixbuf.get_height() / ui_scale

                    render_x = (x + (width / 2) - (logical_width / 2))
                    render_y = (y + (height / 2) - (logical_height / 2))

                    if pixbuf:
                        surface = Gdk.cairo_surface_create_from_pixbuf(pixbuf,
                                                                       ui_scale,
                                                                       self.get_window())

                    cr.set_source_surface(surface,
                                          render_x,
                                          render_y)

                    cr.paint()

                    if self.lockscreen_layout_source.dupe_id > 0:
                        x = render_x + logical_width / 2
                        y = render_y + logical_height / 2
                        width = logical_width / 2 + 2
                        height = logical_height / 2 + 2

                        cr.set_source_rgba(0, 0, 0, 0.5)
                        cr.rectangle(x, y, width, height)
                        cr.fill()

                        cr.set_source_rgba(1.0, 1.0, 1.0, 0.8)
                        cr.rectangle(x + 1, y + 1, width - 2, height - 2)
                        cr.fill()

                        cr.set_source_rgba(0.0, 0.0, 0.0, 1.0)
                        cr.select_font_face("sans", cairo.FontSlant.NORMAL, cairo.FontWeight.BOLD)
                        cr.set_font_size(height - 2.0)

                        dupe_str = str(self.lockscreen_layout_source.dupe_id)

                        ext = cr.text_extents(dupe_str)
                        cr.move_to((x + (width / 2.0) - (ext.width / 2.0)),
                                   (y + (height / 2.0) + (ext.height / 2.0)))
                        cr.show_text(dupe_str)

                    handled = True
                except GLib.Error:
                    pass

        if not handled:
            name = self.lockscreen_layout_source.short_name
            ctx = widget.get_style_context()
            ctx.save()

            ctx.set_state(Gtk.StateFlags.BACKDROP)
            font_size = ctx.get_property("font-size", Gtk.StateFlags.BACKDROP)
            family = ctx.get_property("font-family", Gtk.StateFlags.BACKDROP)
            cr.select_font_face(family[0], cairo.FONT_WEIGHT_NORMAL, cairo.FONT_SLANT_NORMAL)
            cr.set_font_size(font_size)

            (xb, yb, w, h, xa, ya) = cr.text_extents(name)

            # Drop shadow for visibility - 1px, 1px
            cr.set_source_rgba(0, 0, 0, 0.8)
            cr.move_to((x + (width / 2) - (w / 2)) + 1,
                       (y + (height / 2) + (h / 2) + 1))

            cr.show_text(name)

            # Text

            text_color = widget.get_style_context().get_color(Gtk.StateFlags.BACKDROP)

            Gdk.cairo_set_source_rgba(cr, text_color)
            cr.move_to((x + (width / 2) - (w / 2)),
                       (y + (height / 2) + (h / 2)))

            cr.show_text(name)

            ctx.restore()

        return False

    def start_progress(self):
        self.set_progress_pulse_step(0.2)
        trackers.timer_tracker_get().start("auth-progress",
                                           100,
                                           self.pulse)

    def stop_progress(self):
        trackers.timer_tracker_get().cancel("auth-progress")
        self.set_progress_fraction(0.0)

    def pulse(self):
        """
        Periodic callback for the progress bar.
        """
        self.progress_pulse()
        return True

    def on_current_layout_changed(self, cinnamon):
        if not self.cinnamon.has_multiple_keyboard_layouts():
            return

        self.lockscreen_layout_source = self.cinnamon.get_current_layout_source()

        self.grab_focus()
        self.update_layout_icon()

    def on_layout_sources_changed(self, cinnamon):
        self.system_layout_source = self.cinnamon.get_current_layout_source()
        self.lockscreen_layout_source = self.system_layout_source

        if not self.cinnamon.has_multiple_keyboard_layouts():
            return

        self.set_lockscreen_keyboard_layout()

    def on_icon_pressed(self, entry, icon_pos, event):
        if icon_pos == Gtk.EntryIconPosition.PRIMARY:
            self.cinnamon.activate_next_layout()
        elif icon_pos == Gtk.EntryIconPosition.SECONDARY:
            if self.get_input_purpose() == Gtk.InputPurpose.FREE_FORM:
                self.set_visibility(False)
                self.set_input_purpose(Gtk.InputPurpose.PASSWORD)
                self.set_icon_from_icon_name(Gtk.EntryIconPosition.SECONDARY, "cinnamon-screensaver-view-reveal")
            else:
                self.set_visibility(True)
                self.set_input_purpose(Gtk.InputPurpose.FREE_FORM)
                self.set_icon_from_icon_name(Gtk.EntryIconPosition.SECONDARY, "cinnamon-screensaver-view-conceal")
            self.queue_draw()

    def update_layout_icon(self):
        """
        Set an empty icon here so the widget responds to clicks and allocates space for it.
        We'll do the actual flag or whatever in the 'draw' callback.  Setting the icon here
        also ensures a redraw at the correct time to update the flag image.
        """
        self.set_icon_from_icon_name(Gtk.EntryIconPosition.PRIMARY, "screensaver-blank")
        self.set_icon_tooltip_text(Gtk.EntryIconPosition.PRIMARY, self.lockscreen_layout_source.display_name)

    def on_destroy(self, widget, data=None):
        self.stop_progress()

        self.restore_original_layout()

    def set_lockscreen_keyboard_layout(self):
        # If there are multiple keyboard layouts, we want to store
        # the one the user ends up using in the unlock widget, as they'll
        # want to use the same one each time, at least until they change
        # their password.

        saved_index = settings.get_kb_group()
        new_index = 0

        if saved_index == -1:
            new_index = self.system_layout_source.index
            settings.set_kb_group(new_index)
        else:
            new_index = saved_index

        if new_index != self.system_layout_source.index:
            self.cinnamon.activate_layout_index(new_index)

        self.update_layout_icon()

        trackers.con_tracker_get().connect_after(self,
                                                 "draw",
                                                 self.on_draw)

    def restore_original_layout(self):
        """
        Called when the unlock dialog is destroyed, restores
        the group that was active before the screensaver was activated.
        """
        if settings.get_kb_group() != self.lockscreen_layout_source.index:
            settings.set_kb_group(self.lockscreen_layout_source.index)

        self.cinnamon.activate_layout_index(self.system_layout_source.index)

    def grab_focus(self):
        Gtk.Widget.grab_focus(self)

        length = self.get_buffer().get_length()
        self.select_region(length, -1)



#!/usr/bin/python3

from gi.repository import Gio

# Background settings, the only one interested in this is Backgrounds in
# singletons.py, and then only for the "changed" signal - it reloads
# any individual gsettings internally.

bg_settings = Gio.Settings(schema_id="org.cinnamon.desktop.background")

# Screensaver settings - we have no need to listen to changes here,
# They're read anew each time they're used, and while the screensaver
# is active, the user won't be changing them.

ss_settings = Gio.Settings(schema_id="org.cinnamon.desktop.screensaver")
DEFAULT_MESSAGE_KEY = "default-message"
SCREENSAVER_NAME_KEY = "screensaver-name"
CUSTOM_SCREENSAVER_KEY = "custom-screensaver-command"
USER_SWITCH_ENABLED_KEY = "user-switch-enabled"
IDLE_ACTIVATE_KEY = "idle-activation-enabled"
LOCK_ENABLED_KEY = "lock-enabled"
LOCK_DELAY_KEY = "lock-delay"
USE_CUSTOM_FORMAT_KEY = "use-custom-format"
DATE_FORMAT_KEY = "date-format"
TIME_FORMAT_KEY = "time-format"
FONT_DATE_KEY = "font-date"
FONT_MESSAGE_KEY = "font-message"
FONT_TIME_KEY = "font-time"
KB_LAYOUT_KEY = "layout-group"

SHOW_CLOCK_KEY = "show-clock"
SHOW_ALBUMART = "show-album-art"
ALLOW_SHORTCUTS = "allow-keyboard-shortcuts"
ALLOW_MEDIA_CONTROL = "allow-media-control"
SHOW_INFO_PANEL = "show-info-panel"
FLOATING_WIDGETS = "floating-widgets"

# Interface settings - the same logic applies here as above - we don't
# need to listen to changes to these.
if_settings = Gio.Settings(schema_id="org.cinnamon.desktop.interface")
KBD_LAYOUT_SHOW_FLAGS = "keyboard-layout-show-flags"
KBD_LAYOUT_USE_CAPS = "keyboard-layout-use-upper"
KBD_LAYOUT_PREFER_VARIANT = "keyboard-layout-prefer-variant-names"

osk_settings = Gio.Settings(schema_id="org.cinnamon.keyboard")
OSK_TYPE = "keyboard-type"
OSK_SIZE = "keyboard-size"
OSK_ACTIVATION = "activation-mode"

a11y_settings = Gio.Settings(schema_id="org.cinnamon.desktop.a11y.applications")
OSK_A11Y_ENABLED = "screen-keyboard-enabled"

# Every setting has a getter (and setter, rarely).  This is mainly for
# organizational purposes and cleanliness - it's easier to read in the
# main code if you see "settings.get_default_away_message()" than seeing
# "settings.ss_settings.get_string(settings.DEFAULT_MESSAGE_KEY)" or keeping
# instances of GioSettings wherever we need them.

def _check_string(string):
    if string and string != "":
        return string

    return ""

def get_default_away_message():
    msg = ss_settings.get_string(DEFAULT_MESSAGE_KEY)

    return _check_string(msg)

def get_custom_screensaver():
    cmd = ss_settings.get_string(CUSTOM_SCREENSAVER_KEY)

    return _check_string(cmd)

def get_user_switch_enabled():
    return ss_settings.get_boolean(USER_SWITCH_ENABLED_KEY)

def get_idle_activate():
    return ss_settings.get_boolean(IDLE_ACTIVATE_KEY)

def get_idle_lock_enabled():
    return ss_settings.get_boolean(LOCK_ENABLED_KEY)

def get_idle_lock_delay():
    return ss_settings.get_uint(LOCK_DELAY_KEY)

def get_use_custom_format():
    return ss_settings.get_boolean(USE_CUSTOM_FORMAT_KEY)

def get_custom_date_format():
    date_format = ss_settings.get_string(DATE_FORMAT_KEY)

    return _check_string(date_format)

def get_custom_time_format():
    time_format = ss_settings.get_string(TIME_FORMAT_KEY)

    return _check_string(time_format)

def get_date_font():
    date_font = ss_settings.get_string(FONT_DATE_KEY)

    return _check_string(date_font)

def get_message_font():
    message_font = ss_settings.get_string(FONT_MESSAGE_KEY)

    return _check_string(message_font)

def get_time_font():
    time_font = ss_settings.get_string(FONT_TIME_KEY)

    return _check_string(time_font)

def get_show_flags():
    return if_settings.get_boolean(KBD_LAYOUT_SHOW_FLAGS)

def get_show_upper_case_layout():
    return if_settings.get_boolean(KBD_LAYOUT_USE_CAPS)

def get_use_layout_variant_names():
    return if_settings.get_boolean(KBD_LAYOUT_PREFER_VARIANT)

def get_kb_group():
    return ss_settings.get_int(KB_LAYOUT_KEY)

def set_kb_group(group):
    return ss_settings.set_int(KB_LAYOUT_KEY, group)

def get_show_clock():
    return ss_settings.get_boolean(SHOW_CLOCK_KEY)

def get_show_albumart():
    return ss_settings.get_boolean(SHOW_ALBUMART)

def get_allow_shortcuts():
    return ss_settings.get_boolean(ALLOW_SHORTCUTS)

def get_allow_media_control():
    return ss_settings.get_boolean(ALLOW_MEDIA_CONTROL)

def get_show_info_panel():
    return ss_settings.get_boolean(SHOW_INFO_PANEL)

def get_allow_floating():
    return ss_settings.get_boolean(FLOATING_WIDGETS)

def get_osk_type():
    return osk_settings.get_string(OSK_TYPE)

def get_osk_a11y_active():
    return a11y_settings.get_boolean(OSK_A11Y_ENABLED) and osk_settings.get_string(OSK_ACTIVATION) == 'accessible'


#!/usr/bin/python3

from gi.repository import Gio, Gtk

from util import trackers, settings
from baseWindow import BaseWindow
from floating import Floating
from widgets.framedImage import FramedImage

import singletons
import status

class AlbumArt(Floating, BaseWindow):
    """
    AlbumArt

    It is a child of the Stage's GtkOverlay, and its placement is
    controlled by the overlay's child positioning function.

    When not Awake, it positions itself around all monitors
    using a timer which randomizes its halign and valign properties
    as well as its current monitor.
    """
    def __init__(self, initial_monitor=0, low_res=False):
        super(AlbumArt, self).__init__(initial_monitor, Gtk.Align.END, Gtk.Align.CENTER)
        self.get_style_context().add_class("albumart")

        if not settings.get_show_albumart():
            return

        self.watcher = singletons.MediaPlayerWatcher
        self.player = self.watcher.get_best_player()

        self.current_url = None

        self.image = FramedImage(low_res, scale_up=True)
        self.image.show()
        self.image.set_opacity(0.0)
        self.add(self.image)

        trackers.con_tracker_get().connect(self.image,
                                           "surface-changed",
                                           self.on_surface_changed)

        if self.player is not None:
            trackers.con_tracker_get().connect(self.player,
                                               "metadata-changed",
                                               self.on_metadata_changed)
            self.on_metadata_changed(self.player)

    def on_surface_changed(self, image, surface):
        if surface is not None:
            self.image.set_opacity(1.0)
        else:
            self.image.set_opacity(0.0)

    def on_metadata_changed(self, player):
        self.update_image()

    def update_image(self):
        url = self.player.get_albumart_url()

        if self.player.get_identity() == "spotify":
            url = url.replace("open.spotify.com", "i.scdn.co")

        if url == self.current_url:
            return

        self.current_url = url

        if url == "":
            self.image.clear_image()

        f = Gio.File.new_for_uri(url)

        if f.get_uri_scheme() == "file":
            self.image.set_from_path(f.get_path())
        elif f.get_uri_scheme() == "http":
            self.image.set_from_file(f)



#!/usr/bin/python3

from gi.repository import Gio, GObject, GLib

class BaseClient(GObject.GObject):
    """
    The base constructor for all of our generated GDBusProxies.

    This initializes and sets self.proxy, then fires on_client_setup_complete()
    or on_failure(), which the subclasses implement, depending on the outcome.

    These clients are technically one more level of abstraction than would be
    needed in a perfect world, where all dbus proxies work as they should and
    are supported properly by the actual providers they proxy for, but they
    provide a convenient and relatively clean way of implementing workarounds
    and alternate ways of retrieving or calculating values when the interface
    itself is broken in some way (this is common.)

    They also provide a bit of fault tolerance in cases where one or more of
    these providers do not exist on the bus, and we can provide sane default
    values for our widgets.
    """
    def __init__(self, bustype, proxy_class, service, path):
        """
        Asynchronously initialize the GDBusProxy - we'll call
        on_client_setup_complete() or on_failure() depending on this outcome.
        """
        super(BaseClient, self).__init__()

        self.proxy_class = proxy_class
        self.path = path

        self.proxy = None

        self.watch_name_id = Gio.bus_watch_name(bustype,
                                                service,
                                                Gio.BusNameWatcherFlags.NONE,
                                                self._on_appeared,
                                                self.on_failure)

    def _on_appeared(self, connection, name, name_owner, data=None):
        try:
            self.proxy_class.new(connection,
                                 Gio.DBusProxyFlags.NONE,
                                 name,
                                 self.path,
                                 None,
                                 self._on_proxy_ready)
        except GLib.Error:
            self.proxy = None
            self.on_failure()

        Gio.bus_unwatch_name(self.watch_name_id)

    def _on_proxy_ready(self, object, result, data=None):
        self.proxy = self.proxy_class.new_finish(result)

        self.on_client_setup_complete()

    def ensure_proxy_alive(self):
        """
        Use this as a safety check to see if a given proxy is valid
        and owned.
        """
        return self.proxy and self.proxy.get_name_owner() is not None

    def on_client_setup_complete(self):
        """
        Subclasses must implement this - to complete setup after self.proxy is
        successfully initialized.
        """
        print("You need to implement on_client_setup_complete(self) in your real client class")
        raise NotImplementedError

    def on_failure(self, *args):
        """
        Can be implemented by subclasses, but not necessary.  Nothing further is done anyhow
        if on_client_setup_complete() is never called.
        """
        pass



#!/usr/bin/python3

from gi.repository import Gtk
import random
import status

from util import trackers
from util import settings

class PositionInfo():
    def __init__(self, monitor, halign, valign):
        self.monitor = monitor
        self.halign = halign
        self.valign = valign

class Floating:
    def __init__(self, initial_monitor=0, halign=Gtk.Align.CENTER, valign=Gtk.Align.CENTER):
        super(Floating, self).__init__()
        self.awake_position = PositionInfo(initial_monitor, halign, valign)
        self.next_position = self.awake_position
        self.current_monitor = initial_monitor
        self.apply_next_position()

    def start_positioning(self):
        self.apply_next_position()
        self.show()

    def set_next_position(self, monitor, halign, valign):
        self.next_position = PositionInfo(monitor, halign, valign)

    def apply_next_position(self):
        self.current_monitor = self.next_position.monitor
        self.set_halign(self.next_position.halign)
        self.set_valign(self.next_position.valign)

    def set_awake_position(self, monitor):
        self.awake_position.monitor = monitor
        self.next_position = self.awake_position



#!/usr/bin/python3

from gi.repository import Gio, GLib, CScreensaver, GObject
from enum import IntEnum

from dbusdepot.baseClient import BaseClient
from util import trackers

class PlaybackStatus(IntEnum):
    Unknown = 0
    Playing = 1
    Paused = 2
    Stopped = 3

class MprisClient(BaseClient):
    """
    Represents a media player with an mpris dbus interface.

    There can be as many of these as there are players active in the session,
    but we only control the first active one in our list.

    These are instantiated by our MediaPlayerWatcher.
    """
    __gsignals__ = {
        "status-changed": (GObject.SignalFlags.RUN_LAST, None, (int,)),
        "metadata-changed": (GObject.SignalFlags.RUN_LAST, None, ())
    }
    def __init__(self, name, path):
        super(MprisClient, self).__init__(Gio.BusType.SESSION,
                                          CScreensaver.MediaPlayerProxy,
                                          name,
                                          path)

        self.identity = None
        self.metadata = None
        self.album_name = ""
        self.track_name = ""
        self.artist_name = ""
        self.albumart_url = ""

    def on_client_setup_complete(self):
        trackers.con_tracker_get().connect(self.proxy,
                                           "notify::playback-status",
                                           self.on_playback_status_changed)

        trackers.con_tracker_get().connect(self.proxy,
                                           "notify::metadata",
                                           self.on_metadata_changed)

        self.ensure_metadata()

        # This isn't the app-provided MediaPlayer.Identity, but we'd need to
        # set up another proxy and it's not really necessary for what this is
        # used for.
        self.identity = self.proxy.get_name().rpartition(".")[2]

    def get_identity(self):
        return self.identity

    def get_playback_status(self):
        status = PlaybackStatus.Unknown

        if self.ensure_proxy_alive():
            str_prop = self.proxy.get_property("playback-status")

            try:
                status = PlaybackStatus(eval("PlaybackStatus." + str_prop))
            except (ValueError, TypeError, SyntaxError):
                pass

        return status

    def get_can_play_pause(self):
        if self.ensure_proxy_alive():
            return self.proxy.get_property("can-play") or self.proxy.get_property("can-pause")

        return False

    def get_can_control(self):
        if self.ensure_proxy_alive():
            return self.proxy.get_property("can-control")

        return False

    def play_pause(self):
        if self.ensure_proxy_alive():
            self.proxy.call_play_pause()

    def get_can_go_next(self):
        if self.ensure_proxy_alive():
            return self.proxy.get_property("can-go-next")

        return False

    def go_next(self):
        if self.ensure_proxy_alive():
            self.proxy.call_next()

    def get_can_go_previous(self):
        if self.ensure_proxy_alive():
            return self.proxy.get_property("can-go-previous")

        return False

    def go_previous(self):
        if self.ensure_proxy_alive():
            self.proxy.call_previous()

    def get_name(self):
        if self.ensure_proxy_alive():
            return self.proxy.get_name()

        return ""

    def get_track_name(self):
        self.ensure_metadata()

        return self.track_name

    def get_artist_name(self):
        self.ensure_metadata()

        return self.artist_name

    def get_album_name(self):
        self.ensure_metadata()

        return self.album_name

    def get_albumart_url(self):
        self.ensure_metadata()

        return self.albumart_url

    def on_failure(self, *args):
        pass

    def return_best_string(self, item):
        if type(item) == list:
            return ", ".join(item)
        elif type(item) == str:
            return item
        else:
            return ""

    def ensure_metadata(self):
        if not self.metadata:
            self.metadata = self.proxy.get_property("metadata")
            if self.metadata:
                try:
                    self.track_name = self.return_best_string(self.metadata["xesam:title"])
                except KeyError:
                    self.track_name = ""
                try:
                    self.album_name = self.return_best_string(self.metadata["xesam:album"])
                except KeyError:
                    self.album_name = ""
                try:
                    self.artist_name = self.return_best_string(self.metadata["xesam:albumArtist"])
                except KeyError:
                    try:
                        self.artist_name = self.return_best_string(self.metadata["xesam:artist"])
                    except:
                        self.artist_name = ""
                try:
                    self.albumart_url = self.return_best_string(self.metadata["mpris:artUrl"])
                except KeyError:
                    self.albumart_url = ""

    def on_playback_status_changed(self, proxy, pspec, data=None):
        self.emit("status-changed", self.get_playback_status())

    def on_metadata_changed(self, proxy, pspec, data=None):
        self.metadata = None
        self.ensure_metadata()
        self.emit("metadata-changed")

class MediaPlayerWatcher(GObject.Object):
    """
    Media player interfaces are different from our other interfaces.
    There is no common owned name, players export their own unique interface,
    within the org.mpris.MediaPlayer2.* namespace.  This allows multiple
    players to exist on the bus at one time.  It requires us to list all
    interfaces and filter out all but the ones in that namespace.  We then
    create separate Player clients for each one.
    """
    MPRIS_PATH = "/org/mpris/MediaPlayer2"

    def __init__(self):
        """
        Connect to the bus and retrieve a list of interfaces.
        """
        super(MediaPlayerWatcher, self).__init__()

        self.player_clients = []

        try:
            self.dbus_proxy = Gio.DBusProxy.new_for_bus_sync(Gio.BusType.SESSION,
                                                             Gio.DBusProxyFlags.NONE,
                                                             None,
                                                             "org.freedesktop.DBus",
                                                             "/org/freedesktop/DBus",
                                                             "org.freedesktop.DBus",
                                                             None)

            trackers.con_tracker_get().connect(self.dbus_proxy,
                                               "g-signal",
                                               self.on_dbus_proxy_signal)

            self.find_initial_players()
        except GLib.Error:
            self.dbus_proxy = None
            print("Cannot acquire session org.freedesktop.DBus client to watch for media players")

    def on_dbus_proxy_signal(self, proxy, sender, signal, parameters, data=None):
        """
        NameOwnerChanged is called both when a name appears on the bus, as
        well as when it disappears.  The filled parameters tell us whether we've
        gained or lost the player.
        """
        if signal == "NameOwnerChanged":
            if parameters[2] != "":
                self.on_name_acquired(parameters[0])
            else:
                self.on_name_lost(parameters[0])

    def find_initial_players(self):
        self.dbus_proxy.call("ListNames",
                             None,
                             Gio.DBusCallFlags.NONE,
                             -1,
                             None,
                             self.on_names_listed)

    def on_names_listed(self, bus, result, data=None):
        names = bus.call_finish(result)[0]

        for name in names:
            self.on_name_acquired(name)

    def on_name_acquired(self, name):
        """
        Create an mpris client for any discovered interfaces.
        """
        if name.startswith("org.mpris.MediaPlayer2."):
            self.player_clients.append(MprisClient(name, self.MPRIS_PATH))

    def on_name_lost(self, name):
        """
        Remove any clients that disappear off the bus.
        """
        item = None
        for client in self.player_clients:
            if client.get_name() == name:
                item = client
                break

        if item:
            self.player_clients.remove(item)

    def get_best_player(self):
        """
        Find the first player in our list that is playing, then for one
        that *can* be played.  Players that are simply loaded but don't 
        have any playlist queued up should not pass these tests - we have
        only limited control from the lockscreen.
        """
        
        first_idle = None
        
        for client in self.player_clients:
            if client.get_playback_status() == PlaybackStatus.Playing:
                return client

            if first_idle is None and client.get_can_play_pause() and client.get_can_control():
                first_idle = client

        return first_idle

    def get_all_player_names(self):
        """
        Return a list of all player simple names - this is used by our
        notification code to ignore notifications sent from media players,
        which are usually unimportant, but not marked as transient (which
        would normally cause them to be ignored in the CsNotificationWatcher.)
        """
        ret = []

        for client in self.player_clients:
            fullname = client.get_name()
            split = fullname.split(".")
            ret.append(split[len(split) - 1].lower())

        return ret



#!/usr/bin/python3

import gi
gi.require_version('CScreensaver', '1.0')
from gi.repository import GLib, Gio, GObject, CScreensaver

# TODO
# self.monitors, etc.. replace or at least prefer this over CsScreen, as it will be more accurate.
# Nothing currently listens to muffin-config-changed. This class is only used to initialize the event filters.

class MuffinClient(GObject.Object):
    MUFFIN_SERVICE = "org.cinnamon.Muffin.DisplayConfig"
    MUFFIN_PATH    = "/org/cinnamon/Muffin/DisplayConfig"

    __gsignals__ = {
        'muffin-config-changed': (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self):
        GObject.Object.__init__(self)

        self.proxy = None
        self.using_fractional_scaling = False

        try:
            self.proxy = CScreensaver.MuffinDisplayConfigProxy.new_for_bus_sync(Gio.BusType.SESSION,
                                                                                Gio.DBusProxyFlags.DO_NOT_LOAD_PROPERTIES |
                                                                                  Gio.DBusProxyFlags.DO_NOT_AUTO_START,
                                                                                self.MUFFIN_SERVICE,
                                                                                self.MUFFIN_PATH,
                                                                                None)
            self.proxy.connect("monitors-changed", self.on_monitors_changed)
            # cinnamon restart (monitors-changed isn't emitted at muffin startup)
            self.proxy.connect("notify::g-name-owner", self.on_name_owner_changed)
            self.update()
        except GLib.Error as e:
            print(f"Could not connect to Muffin's DisplayConfig service: {e}", flush=True)

    def on_monitors_changed(self, proxy):
        self.update()

    def on_name_owner_changed(self, proxy, pspec):
        if proxy.get_name_owner() is not None:
            self.update()

    def update(self):
        if self.read_current_state():
            self.emit("muffin-config-changed")

    def read_current_state(self, *args):
        old_scaling = self.using_fractional_scaling

        if self.proxy.get_name_owner() is None:
            print("Muffin not running, skipping fractional scaling check.")
            return False

        try:
            logical_monitors = self.proxy.call_get_current_state_sync(None)[2]
        except GLib.Error as e:
            print(f"Could not read current state from Muffin: {e}", flush=True)
            self.using_fractional_scaling = False
            return self.using_fractional_scaling != old_scaling

        fractional = False
        previous_scale = -1

        for monitor in logical_monitors.unpack():
            scale = monitor[2]

            # one or more monitors using some non-integer scale.
            if int(scale) != scale:
                fractional = True
                break

            # multiple monitors with non-identical scales
            if previous_scale > 0 and scale != previous_scale:
                fractional = True
                break

            previous_scale = scale

        self.using_fractional_scaling = fractional
        print(f"Fractional scaling active: {self.using_fractional_scaling}", flush=True)
        return self.using_fractional_scaling != old_scaling

    def get_using_fractional_scaling(self):
        return self.using_fractional_scaling



#!/usr/bin/python3

import gi
gi.require_version('CDesktopEnums', '3.0')

from gi.repository import GLib, Gtk, Gdk, CScreensaver, CDesktopEnums, GObject
import random

import status
import constants as c
import singletons
from floating import Floating
from monitorView import MonitorView
from unlock import UnlockDialog
from clock import ClockWidget
from albumArt import AlbumArt
from audioPanel import AudioPanel
from infoPanel import InfoPanel
from util import utils, trackers, settings
from util.eventHandler import EventHandler
from util.utils import DEBUG

try:
    from osk import OnScreenKeyboard
except Exception as e:
    print("Failed to import on-screen keyboard: %s" % str(e))
    OnScreenKeyboard = None

FLOATER_POSITIONING_TIMEOUT = 30

class Stage(Gtk.Window):
    """
    The Stage is the toplevel window of the entire screensaver while
    in Active mode.

    It's the first thing made, the last thing destroyed, and all other
    widgets live inside of it (or rather, inside the GtkOverlay below)

    It is Gtk.WindowType.POPUP to avoid being managed/composited by muffin,
    and to prevent animation during its creation and destruction.

    The Stage responds pretty much only to the instructions of the
    ScreensaverManager.
    """
    __gsignals__ = {
        'needs-refresh': (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self, manager, away_message):
        if status.InteractiveDebug:
            Gtk.Window.__init__(self,
                                type=Gtk.WindowType.TOPLEVEL,
                                decorated=True,
                                skip_taskbar_hint=False)
        else:
            Gtk.Window.__init__(self,
                                type=Gtk.WindowType.POPUP,
                                decorated=False,
                                skip_taskbar_hint=True)

        self.get_style_context().add_class("csstage")

        trackers.con_tracker_get().connect(singletons.Backgrounds,
                                           "changed",
                                           self.on_bg_changed)

        self.destroying = False

        self.manager = manager
        self.away_message = away_message

        self.activate_callback = None

        self.monitors = []
        self.last_focus_monitor = -1
        self.overlay = None
        self.clock_widget = None
        self.albumart_widget = None
        self.unlock_dialog = None
        self.audio_panel = None
        self.info_panel = None
        self.osk = None

        self.floaters = []
        self.floaters_need_update = False

        self.event_handler = EventHandler(manager)

        self.get_style_context().remove_class("background")

        self.set_events(self.get_events() |
                        Gdk.EventMask.POINTER_MOTION_MASK |
                        Gdk.EventMask.BUTTON_PRESS_MASK |
                        Gdk.EventMask.BUTTON_RELEASE_MASK |
                        Gdk.EventMask.KEY_PRESS_MASK |
                        Gdk.EventMask.KEY_RELEASE_MASK |
                        Gdk.EventMask.EXPOSURE_MASK |
                        Gdk.EventMask.VISIBILITY_NOTIFY_MASK |
                        Gdk.EventMask.ENTER_NOTIFY_MASK |
                        Gdk.EventMask.LEAVE_NOTIFY_MASK |
                        Gdk.EventMask.FOCUS_CHANGE_MASK)

        c = Gdk.RGBA(0, 0, 0, 0)
        self.override_background_color(Gtk.StateFlags.NORMAL, c)

        self.update_geometry()

        self.overlay = Gtk.Overlay()

        trackers.con_tracker_get().connect(self.overlay,
                                           "realize",
                                           self.on_realized)

        trackers.con_tracker_get().connect(self.overlay,
                                           "get-child-position",
                                           self.position_overlay_child)

        self.overlay.show_all()
        self.add(self.overlay)

        # We hang onto the UPowerClient here so power events can
        # trigger changes to the info panel.
        self.power_client = singletons.UPowerClient

        trackers.con_tracker_get().connect(self.power_client,
                                           "power-state-changed",
                                           self.on_power_state_changed)

        # This filter suppresses any other windows that might share
        # our window group in muffin, from showing up over the Stage.
        # For instance: Chrome and Firefox native notifications.
        self.gdk_filter = CScreensaver.GdkEventFilter.new(self, 0)

        trackers.con_tracker_get().connect(status.screen,
                                           "size-changed",
                                           self.on_screen_size_changed)

        trackers.con_tracker_get().connect(status.screen,
                                           "monitors-changed",
                                           self.on_monitors_changed)

        trackers.con_tracker_get().connect(status.screen,
                                           "composited-changed",
                                           self.on_composited_changed)

        trackers.con_tracker_get().connect(self,
                                           "grab-broken-event",
                                           self.on_grab_broken_event)

        if status.InteractiveDebug:
            self.set_interactive_debugging(True)

    def update_monitors(self):
        self.destroy_monitor_views()

        try:
            self.setup_monitors()
            for monitor in self.monitors:
                self.sink_child_widget(monitor)
        except Exception as e:
            print("Problem updating monitor views views: %s" % str(e))

    def on_screen_size_changed(self, screen, data=None):
        """
        The screen changing size should be acted upon immediately, to ensure coverage.
        Wallpapers are secondary.
        """

        DEBUG("Stage: Received screen size-changed signal, refreshing stage")

        self.emit("needs-refresh")

    def on_monitors_changed(self, screen, data=None):
        """
        Updating monitors also will trigger an immediate stage coverage update (same
        as on_screen_size_changed), and follow up at idle with actual monitor view
        refreshes (wallpapers.)
        """
        DEBUG("Stage: Received screen monitors-changed signal, refreshing stage")

        self.emit("needs-refresh")

    def on_composited_changed(self, screen, data=None):
        if self.get_realized():
            DEBUG("Stage: Received screen composited-changed signal, refreshing stage")

            self.emit("needs-refresh")

    def on_grab_broken_event(self, widget, event, data=None):
        GObject.idle_add(self.manager.grab_stage)

        return False

    def activate(self, callback):
        """
        This is the primary way of making the Stage visible.
        """

        self.set_opacity(1.0)
        self.move_onscreen()
        self.show()

        if self.get_realized():
            callback()
        else:
            self.activate_callback = callback

    def deactivate(self, callback):
        """
        This is the primary way of destroying the stage.
        """
        self.hide()
        callback()

    def on_realized(self, widget):
        """
        Repositions the window when it is realized, to cover the entire
        GdkScreen (a rectangle exactly encompassing all monitors.)

        From here we also proceed to construct all overlay children and
        activate our window suppressor.
        """
        window = self.get_window()
        utils.override_user_time(window)

        CScreensaver.Screen.set_net_wm_name(window, "cinnamon-screensaver-window")

        self.setup_children()

        self.gdk_filter.start(singletons.MuffinClient.get_using_fractional_scaling(), status.Debug)

        trackers.con_tracker_get().disconnect(self.overlay,
                                              "realize",
                                              self.on_realized)

        if self.activate_callback is not None:
            self.activate_callback()

    def move_onscreen(self):
        w = self.get_window()

        if w:
            w.move_resize(self.rect.x,
                          self.rect.y,
                          self.rect.width,
                          self.rect.height)

        self.move(self.rect.x, self.rect.y)
        self.resize(self.rect.width, self.rect.height)

    def deactivate_after_timeout(self):
        self.manager.set_active(False)

    def setup_children(self):
        """
        Creates all of our overlay children.  If a new 'widget' gets added,
        this should be the setup point for it.

        We bail if something goes wrong on a critical widget - a monitor view or
        unlock widget.
        """
        total_failure = False

        try:
            self.setup_monitors()
        except Exception as e:
            print("Problem setting up monitor views: %s" % str(e))
            total_failure = True

        try:
            self.setup_unlock()
        except Exception as e:
            print("Problem setting up unlock dialog: %s" % str(e))
            total_failure = True

        if not total_failure:
            try:
                self.setup_clock()
            except Exception as e:
                print("Problem setting up clock widget: %s" % str(e))
                self.clock_widget = None

            if OnScreenKeyboard:
                try:
                    self.setup_osk()
                except Exception as e:
                    print("Problem setting up on-screen keyboard: %s" % str(e))
                    self.osk = None
            else:
                self.osk = None

            trackers.timer_tracker_get().start("setup-delayed-components",
                                               2000,
                                               self.setup_delayed_components)
        else:
            print("Total failure somewhere, deactivating screensaver.")
            GObject.idle_add(self.deactivate_after_timeout)

    def setup_delayed_components(self, data=None):
        try:
            self.setup_albumart()
        except Exception as e:
            print("Problem setting up albumart widget: %s" % str(e))
            self.albumart_widget = None
        try:
            self.setup_status_bars()
        except Exception as e:
            print("Problem setting up status bars: %s" % str(e))
            self.audio_panel = None
            self.info_panel = None

        self.start_float_timer()

    def start_float_timer(self):
        if settings.get_allow_floating():
            if status.Debug:
                timeout = 5
            else:
                timeout = FLOATER_POSITIONING_TIMEOUT
            trackers.timer_tracker_get().start_seconds("floater-update", timeout, self.update_floaters)

    def update_floaters(self):
        self.floaters_need_update = True
        self.overlay.queue_allocate()
        return GLib.SOURCE_CONTINUE

    def stop_float_timer(self):
        trackers.timer_tracker_get().cancel("floater-update")
        self.floaters_need_update = False

    def destroy_children(self):
        self.stop_float_timer()

        try:
            self.destroy_monitor_views()
        except Exception as e:
            print(e)

        try:
            if self.unlock_dialog is not None:
                self.unlock_dialog.destroy()
        except Exception as e:
            print(e)

        try:
            if self.clock_widget is not None:
                self.clock_widget.destroy()
        except Exception as e:
            print(e)

        try:
            if self.albumart_widget is not None:
                self.albumart_widget.destroy()
        except Exception as e:
            print(e)

        try:
            if self.info_panel is not None:
                self.info_panel.destroy()
        except Exception as e:
            print(e)

        try:
            if self.info_panel is not None:
                self.audio_panel.destroy()
        except Exception as e:
            print(e)

        try:
            if self.osk is not None:
                self.osk.destroy()
        except Exception as e:
            print(e)

        self.unlock_dialog = None
        self.clock_widget = None
        self.albumart_widget = None
        self.info_panel = None
        self.audio_panel = None
        self.osk = None
        self.away_message = None

        self.monitors = []
        self.floaters = []

    def destroy_stage(self):
        """
        Performs all tear-down necessary to destroy the Stage, destroying
        all children in the process, and finally destroying itself.
        """

        trackers.con_tracker_get().disconnect(self.unlock_dialog,
                                              "inhibit-timeout",
                                              self.set_timeout_active)

        trackers.con_tracker_get().disconnect(self.unlock_dialog,
                                              "uninhibit-timeout",
                                              self.set_timeout_active)

        trackers.con_tracker_get().disconnect(self.unlock_dialog,
                                              "authenticate-success",
                                              self.authentication_result_callback)

        trackers.con_tracker_get().disconnect(self.unlock_dialog,
                                              "authenticate-failure",
                                              self.authentication_result_callback)

        trackers.con_tracker_get().disconnect(self.unlock_dialog,
                                              "authenticate-cancel",
                                              self.authentication_cancel_callback)

        trackers.con_tracker_get().disconnect(singletons.Backgrounds,
                                              "changed",
                                              self.on_bg_changed)

        trackers.con_tracker_get().disconnect(self.power_client,
                                              "power-state-changed",
                                              self.on_power_state_changed)

        trackers.con_tracker_get().disconnect(self,
                                              "grab-broken-event",
                                              self.on_grab_broken_event)

        self.set_timeout_active(None, False)

        trackers.timer_tracker_get().cancel("setup-delayed-components")
        self.destroy_children()

        self.gdk_filter.stop()
        self.gdk_filter = None

        trackers.con_tracker_get().disconnect(status.screen,
                                              "size-changed",
                                              self.on_screen_size_changed)

        trackers.con_tracker_get().disconnect(status.screen,
                                              "monitors-changed",
                                              self.on_monitors_changed)

        trackers.con_tracker_get().disconnect(self.overlay,
                                              "get-child-position",
                                              self.position_overlay_child)

        self.destroy()

    def setup_monitors(self):
        """
        Iterate through the monitors, and create MonitorViews for each one
        to cover them.
        """
        self.monitors = []
        status.Spanned = settings.bg_settings.get_enum("picture-options") == CDesktopEnums.BackgroundStyle.SPANNED

        if status.InteractiveDebug or status.Spanned:
            monitors = (status.screen.get_primary_monitor(),)
        else:
            n = status.screen.get_n_monitors()
            monitors = ()
            for i in range(n):
                monitors += (i,)

        for index in monitors:
            monitor = MonitorView(index)

            image = Gtk.Image()

            singletons.Backgrounds.create_and_set_gtk_image (image,
                                                             monitor.rect.width,
                                                             monitor.rect.height)

            monitor.set_next_wallpaper_image(image)

            self.monitors.append(monitor)

            self.add_child_widget(monitor)

        self.update_monitor_views()

    def on_bg_changed(self, bg):
        """
        Callback for our GnomeBackground instance, this tells us when
        the background settings have changed, so we can update our wallpaper.
        """
        for monitor in self.monitors:
            image = Gtk.Image()

            singletons.Backgrounds.create_and_set_gtk_image (image,
                                                             monitor.rect.width,
                                                             monitor.rect.height)

            monitor.set_next_wallpaper_image(image)

    def on_power_state_changed(self, client, data=None):
        """
        Callback for UPower changes, this will make our MonitorViews update
        themselves according to user setting and power state.
        """
        DEBUG("stage: Power state changed, updating info panel")

        if self.info_panel is not None:
            self.info_panel.update_visibility()

    def setup_clock(self):
        """
        Construct the clock widget and add it to the overlay, but only actually
        show it if we're a) Not running a plug-in, and b) The user wants it via
        preferences.

        Initially invisible, regardless - its visibility is controlled via its
        own positioning timer.
        """
        self.clock_widget = ClockWidget(self.away_message, status.screen.get_mouse_monitor(), status.screen.get_low_res_mode())
        self.add_child_widget(self.clock_widget)

        self.floaters.append(self.clock_widget)

        if settings.get_show_clock():
            self.clock_widget.start_positioning()

    def setup_albumart(self):
        """
        Construct the AlbumArt widget and add it to the overlay, but only actually
        show it if we're a) Not running a plug-in, and b) The user wants it via
        preferences.

        Initially invisible, regardless - its visibility is controlled via its
        own positioning timer.
        """
        self.albumart_widget = AlbumArt(status.screen.get_mouse_monitor(), status.screen.get_low_res_mode())
        self.add_child_widget(self.albumart_widget)

        self.floaters.append(self.albumart_widget)

        if settings.get_show_albumart():
            self.albumart_widget.start_positioning()

    def setup_osk(self):
        self.osk = OnScreenKeyboard()

        self.add_child_widget(self.osk)

    def setup_unlock(self):
        """
        Construct the unlock dialog widget and add it to the overlay.  It will always
        initially be invisible.

        Any time the screensaver is awake, and the unlock dialog is raised, a timer runs.
        After a certain elapsed time, the state will be reset, and the dialog will be hidden
        once more.  Mouse and key events reset this timer, and the act of authentication
        temporarily suspends it - the unlock widget accomplishes this via its inhibit- and
        uninhibit-timeout signals

        We also listen to actual authentication events, to destroy the stage if there is success,
        and to do something cute if we fail (for now, this consists of 'blinking' the unlock
        dialog.)
        """
        self.unlock_dialog = UnlockDialog()
        self.set_default(self.unlock_dialog.auth_unlock_button)
        self.add_child_widget(self.unlock_dialog)

        # Prevent a dialog timeout during authentication
        trackers.con_tracker_get().connect(self.unlock_dialog,
                                           "inhibit-timeout",
                                           self.set_timeout_active, False)
        trackers.con_tracker_get().connect(self.unlock_dialog,
                                           "uninhibit-timeout",
                                           self.set_timeout_active, True)

        # Respond to authentication success/failure
        trackers.con_tracker_get().connect(self.unlock_dialog,
                                           "authenticate-success",
                                           self.authentication_result_callback, True)
        trackers.con_tracker_get().connect(self.unlock_dialog,
                                           "authenticate-failure",
                                           self.authentication_result_callback, False)
        trackers.con_tracker_get().connect(self.unlock_dialog,
                                           "authenticate-cancel",
                                           self.authentication_cancel_callback)

    def setup_status_bars(self):
        """
        Constructs the AudioPanel and InfoPanel and adds them to the overlay.
        """
        self.audio_panel = AudioPanel()
        self.add_child_widget(self.audio_panel)

        if status.Awake:
            self.audio_panel.show_panel()

        self.info_panel = InfoPanel()
        self.add_child_widget(self.info_panel)

        self.info_panel.refresh_power_state()

    def queue_dialog_key_event(self, event):
        """
        Sent from our EventHandler via the ScreensaverManager, this catches
        initial key events before the unlock dialog is made visible, so that
        the user doesn't have to first jiggle the mouse to wake things up before
        beginning to type their password.  They can just start typing, and no
        keystrokes will be lost.
        """
        self.unlock_dialog.queue_key_event(event)

# Timer stuff - after a certain time, the unlock dialog will cancel itself.
# This timer is suspended during authentication, and any time a new user event is received

    def reset_timeout(self):
        """
        This is called when any user event is received in our EventHandler.
        This restarts our dialog timeout.
        """
        self.set_timeout_active(None, True)

    def set_timeout_active(self, dialog, active):
        """
        Start or stop the dialog timer
        """
        if active and not status.InteractiveDebug:
            trackers.timer_tracker_get().start("wake-timeout",
                                               c.UNLOCK_TIMEOUT * 1000,
                                               self.on_wake_timeout)
        else:
            trackers.timer_tracker_get().cancel("wake-timeout")

    def on_wake_timeout(self):
        """
        Go back to Sleep if we hit our timer limit
        """
        self.set_timeout_active(None, False)
        self.manager.cancel_unlocking()

        return False

    def authentication_result_callback(self, dialog, success):
        """
        Called by authentication success or failure.  Either starts
        the stage despawning process or simply 'blinks' the unlock
        widget, depending on the outcome.
        """
        if success:
            if self.clock_widget is not None:
                self.clock_widget.hide()
            if self.albumart_widget is not None:
                self.albumart_widget.hide()
            self.unlock_dialog.hide()
            self.manager.unlock()
        else:
            self.unlock_dialog.blink()

    def authentication_cancel_callback(self, dialog):
        self.manager.cancel_unlocking()

    def set_message(self, msg):
        """
        Passes along an away-message to the clock.
        """
        if self.clock_widget is not None:
            self.clock_widget.set_message(msg)

    def initialize_pam(self):
        return self.unlock_dialog.initialize_auth_client()

    def raise_unlock_widget(self):
        """
        Bring the unlock widget to the front and make sure it's visible.
        """
        self.reset_timeout()

        if status.Awake:
            return

        status.screen.place_pointer_in_primary_monitor ()

        utils.clear_clipboards(self.unlock_dialog)

        self.stop_float_timer()

        status.Awake = True

        if self.info_panel:
            self.info_panel.refresh_power_state()

        self.unlock_dialog.show()

        if self.audio_panel is not None:
            self.audio_panel.show_panel()
        if self.info_panel is not None:
            self.info_panel.refresh_power_state()
        if self.osk is not None:
            self.osk.show()

    def cancel_unlocking(self):
        if self.unlock_dialog:
            self.unlock_dialog.cancel()

        self.set_timeout_active(None, False)
        utils.clear_clipboards(self.unlock_dialog)

        if self.unlock_dialog is not None:
            self.unlock_dialog.hide()
        if self.audio_panel is not None:
            self.audio_panel.hide()
        if self.info_panel is not None:
            self.info_panel.hide()
        if self.osk is not None:
            self.osk.hide()

        status.Awake = False

        self.update_monitor_views()
        self.start_float_timer()

        if self.info_panel is not None:
            self.info_panel.refresh_power_state()

    def update_monitor_views(self):
        """
        Updates all of our MonitorViews based on the power
        or Awake states.
        """
        for monitor in self.monitors:
                monitor.show()

    def destroy_monitor_views(self):
        """
        Destroy all MonitorViews
        """
        for monitor in self.monitors:
            monitor.destroy()
            del monitor

    def do_motion_notify_event(self, event):
        """
        GtkWidget class motion-event handler.  Delegate to EventHandler
        """
        return self.event_handler.on_motion_event(event)

    def do_key_press_event(self, event):
        """
        GtkWidget class key-press-event handler.  Delegate to EventHandler
        """
        return self.event_handler.on_key_press_event(event)

    def do_button_press_event(self, event):
        """
        GtkWidget class button-press-event handler.  Delegate to EventHandler
        """
        return self.event_handler.on_button_press_event(event)

    def update_geometry(self):
        """
        Override BaseWindow.update_geometry() - the Stage should always be the
        GdkScreen size, unless status.InteractiveDebug is True
        """

        if status.InteractiveDebug:
            monitor_n = status.screen.get_primary_monitor()
            self.rect = status.screen.get_monitor_geometry(monitor_n)
        else:
            self.rect = status.screen.get_screen_geometry()

            scale = status.screen.get_global_scale()
            self.rect.width *= scale
            self.rect.height *= scale

        DEBUG("Stage.update_geometry - new backdrop position: %d, %d  new size: %d x %d" % (self.rect.x, self.rect.y, self.rect.width, self.rect.height))

        hints = Gdk.Geometry()
        hints.min_width = self.rect.width
        hints.min_height = self.rect.height
        hints.max_width = self.rect.width
        hints.max_height = self.rect.height
        hints.base_width = self.rect.width
        hints.base_height = self.rect.height

        self.set_geometry_hints(self, hints, Gdk.WindowHints.MIN_SIZE | Gdk.WindowHints.MAX_SIZE | Gdk.WindowHints.BASE_SIZE)

# Overlay window management

    def get_mouse_monitor(self):
        if status.InteractiveDebug:
            return status.screen.get_primary_monitor()
        else:
            return status.screen.get_mouse_monitor()

    def maybe_update_layout(self):
        """
        Called on all user events, moves widgets to the currently
        focused monitor if it changes (whichever monitor the mouse is in)
        """
        current_focus_monitor = status.screen.get_mouse_monitor()

        if self.last_focus_monitor == -1:
            self.last_focus_monitor = current_focus_monitor
            return

        if self.unlock_dialog and current_focus_monitor != self.last_focus_monitor:
            self.last_focus_monitor = current_focus_monitor
            self.overlay.queue_resize()

    def add_child_widget(self, widget):
        """
        Add a new child to the overlay
        """
        self.overlay.add_overlay(widget)

    def sink_child_widget(self, widget):
        """
        Move a child to the bottom of the overlay
        """
        self.overlay.reorder_overlay(widget, 0)

    def position_overlay_child(self, overlay, child, allocation):
        """
        Callback for our GtkOverlay, think of this as a mini-
        window manager for our Stage.

        Depending on what type child is, we position it differently.
        We always call child.get_preferred_size() whether we plan to use
        it or not - this prevents allocation warning spew, particularly in
        Gtk >= 3.20.

        Returning True says, yes draw it.  Returning False tells it to skip
        drawing.

        If a new widget type is introduced that spawns directly on the stage,
        it must have its own handling code here.
        """
        if isinstance(child, MonitorView):
            """
            MonitorView is always the size and position of its assigned monitor.
            This is calculated and stored by the child in child.rect)
            """
            w, h = child.get_preferred_size()
            allocation.x = child.rect.x
            allocation.y = child.rect.y
            allocation.width = child.rect.width
            allocation.height = child.rect.height

            return True

        if isinstance(child, UnlockDialog):
            """
            UnlockDialog always shows on the currently focused monitor (the one the
            mouse is currently in), and is kept centered.
            """
            monitor = status.screen.get_mouse_monitor()
            monitor_rect = status.screen.get_monitor_geometry(monitor)

            min_rect, nat_rect = child.get_preferred_size()

            allocation.width = nat_rect.width
            allocation.height = nat_rect.height

            allocation.x = monitor_rect.x + (monitor_rect.width / 2) - (allocation.width / 2)
            allocation.y = monitor_rect.y + (monitor_rect.height / 2) - (allocation.height / 2)

            return True

        if isinstance(child, Floating):
            """
            ClockWidget and AlbumArt behave differently depending on if status.Awake is True or not.

            The widgets' halign and valign properties are used to store their gross position on the
            monitor.  This limits the number of possible positions to (3 * 3 * n_monitors) when our
            screensaver is not Awake, and the widgets have an internal timer that randomizes halign,
            valign, and current monitor every so many seconds, calling a queue_resize on itself after
            each timer tick (which forces this function to run).
            """
            if settings.get_allow_floating():
                if self.floaters_need_update:
                    alignments = [int(Gtk.Align.START), int(Gtk.Align.END), int(Gtk.Align.CENTER)]
                    n_monitors = status.screen.get_n_monitors()
                    alignment_positions = []
                    for i in range(n_monitors):
                        for halign in alignments:
                            for valign in alignments:
                                alignment_positions.append((i, halign, valign))

                    for floater in self.floaters:
                        position = alignment_positions.pop(random.randint(0, len(alignment_positions) - 1))
                        floater.set_next_position(*position)

                    self.floaters_need_update = False

                child.apply_next_position()

            min_rect, nat_rect = child.get_preferred_size()
            current_monitor = child.current_monitor
            monitor_rect = status.screen.get_monitor_geometry(current_monitor)
            region_w = monitor_rect.width / 3
            region_h = monitor_rect.height / 3

            if status.Awake:
                current_monitor = status.screen.get_mouse_monitor()
                monitor_rect = status.screen.get_monitor_geometry(current_monitor)
                region_w = monitor_rect.width / 3
                region_h = monitor_rect.height / 3

                """
                If we're Awake, force the clock to track to the active monitor, and be aligned to
                the left-center.  The albumart widget aligns right-center.
                """
                unlock_mw, unlock_nw = self.unlock_dialog.get_preferred_width()
                """
                If, for whatever reason, we need more than 1/3 of the screen to fully display
                the unlock dialog, reduce our available region width to accommodate it, reducing
                the allocation for the floating widgets as required.
                """
                if unlock_nw > region_w:
                    region_w = (monitor_rect.width - unlock_nw) / 2

                child.set_awake_position(current_monitor)
                child.apply_next_position()

            # Restrict the widget size to the allowable region sizes if necessary.
            allocation.width = min(nat_rect.width, region_w)
            allocation.height = min(nat_rect.height, region_h)

            # Calculate padding required to center widgets within their particular 1/9th of the monitor
            padding_left = padding_right = (region_w - allocation.width) / 2
            padding_top = padding_bottom = (region_h - allocation.height) / 2

            halign = child.get_halign()
            valign = child.get_valign()

            if halign == Gtk.Align.START:
                allocation.x = monitor_rect.x + padding_left
            elif halign == Gtk.Align.CENTER:
                allocation.x = monitor_rect.x + (monitor_rect.width / 2) - (allocation.width / 2)
            elif halign == Gtk.Align.END:
                allocation.x = monitor_rect.x + monitor_rect.width - allocation.width - padding_right

            if valign == Gtk.Align.START:
                allocation.y = monitor_rect.y + padding_top
            elif valign == Gtk.Align.CENTER:
                allocation.y = monitor_rect.y + (monitor_rect.height / 2) - (allocation.height / 2)
            elif valign == Gtk.Align.END:
                allocation.y = monitor_rect.y + monitor_rect.height - allocation.height - padding_bottom

            return True

        if isinstance(child, AudioPanel):
            """
            The AudioPanel is only shown when Awake, and attaches
            itself to the upper-left corner of the active monitor.
            """
            min_rect, nat_rect = child.get_preferred_size()

            if status.Awake:
                current_monitor = status.screen.get_mouse_monitor()
                monitor_rect = status.screen.get_monitor_geometry(current_monitor)
                allocation.x = monitor_rect.x
                allocation.y = monitor_rect.y
                allocation.width = nat_rect.width
                allocation.height = nat_rect.height
            else:
                allocation.x = child.rect.x
                allocation.y = child.rect.y
                allocation.width = nat_rect.width
                allocation.height = nat_rect.height

            return True

        if isinstance(child, InfoPanel):
            """
            The InfoPanel can be shown while not Awake, but will only appear if a) We have received
            notifications while the screensaver is running, or b) we're either on battery
            or plugged in but with a non-full battery.  It attaches itself to the upper-right
            corner of the monitor.
            """
            min_rect, nat_rect = child.get_preferred_size()

            if status.Awake:
                current_monitor = status.screen.get_mouse_monitor()
                monitor_rect = status.screen.get_monitor_geometry(current_monitor)
                allocation.x = monitor_rect.x + monitor_rect.width - nat_rect.width
                allocation.y = monitor_rect.y
                allocation.width = nat_rect.width
                allocation.height = nat_rect.height
            else:
                allocation.x = child.rect.x + child.rect.width - nat_rect.width
                allocation.y = child.rect.y
                allocation.width = nat_rect.width
                allocation.height = nat_rect.height

            return True

        if isinstance(child, OnScreenKeyboard):
            """
            The InfoPanel can be shown while not Awake, but will only appear if a) We have received
            notifications while the screensaver is running, or b) we're either on battery
            or plugged in but with a non-full battery.  It attaches itself to the upper-right
            corner of the monitor.
            """
            min_rect, nat_rect = child.get_preferred_size()

            current_monitor = status.screen.get_mouse_monitor()
            monitor_rect = status.screen.get_monitor_geometry(current_monitor)
            allocation.x = monitor_rect.x
            allocation.y = monitor_rect.y + monitor_rect.height - (monitor_rect.height / 3)
            allocation.width = monitor_rect.width
            allocation.height = monitor_rect.height / 3

            return True

        return False



#!/usr/bin/python3

from gi.repository import Gio, CScreensaver, GObject

from dbusdepot.baseClient import BaseClient
from util import trackers

# see cinnamon/files/usr/share/cinnamon/cinnamon-settings/bin/InputSources.py
class CurrentInputSource:
    def __init__(self, source):
        self.type, self.id, self.index,         \
            self.display_name, self.short_name, \
            self.flag_name, self.xkbid,         \
            self.xkb_layout, self.xkb_variant,  \
            self.preferences,                   \
            self.dupe_id, self.active           \
                 = source

class CinnamonClient(BaseClient):
    """
    Client to talk to Cinnamon's dbus interface.

    Used to deactivate special modal cinnamon states and deal with
    keyboard layout info.
    """
    CINNAMON_SERVICE = "org.Cinnamon"
    CINNAMON_PATH    = "/org/Cinnamon"
    __gsignals__ = {
        'current-input-source-changed': (GObject.SignalFlags.RUN_LAST, None, ()),
        'input-sources-changed': (GObject.SignalFlags.RUN_LAST, None, ())
    }
    def __init__(self):
        super(CinnamonClient, self).__init__(Gio.BusType.SESSION,
                                             CScreensaver.CinnamonProxy,
                                             self.CINNAMON_SERVICE,
                                             self.CINNAMON_PATH)
        self.sources = []

    def on_client_setup_complete(self):
        trackers.con_tracker_get().connect(self.proxy, "g-signal", self.on_cinnamon_signal)
        self.update_layout_sources()

    def on_cinnamon_signal(self, proxy, sender, signal, params, data=None):
        if signal == "CurrentInputSourceChanged":
            self.update_current_layout(params[0])
        elif signal == "InputSourcesChanged":
            self.update_layout_sources()

    def update_layout_sources(self):
        self.proxy.GetInputSources(result_handler=self.get_input_sources_callback,
                                   error_handler=self.get_input_sources_error)

    def get_input_sources_callback(self, proxy, sources, data=None):
        self.sources = []

        for source in sources:
            input_source = CurrentInputSource(source)
            if input_source.type == "xkb":
                self.sources.append(input_source)
        self.emit("input-sources-changed")

    def get_input_sources_error(self, proxy, error, data=None):
        print("Failed to get keyboard layouts from Cinnamon - multiple layouts will not be available: %s" % error.message)

    def has_multiple_keyboard_layouts(self):
        return len(self.sources) > 1

    def update_current_layout(self, layout):
        for source in self.sources:
            source.active = source.id == layout
        self.emit("current-input-source-changed")

    def get_current_layout_source(self):
        for source in self.sources:
            if source.active:
                return source
        return None

    def activate_layout_index(self, index):
        self.proxy.ActivateInputSourceIndex("(i)", index)

    def activate_next_layout(self):
        current = 0

        for i in range(0, len(self.sources)):
            source = self.sources[i]
            if source.active:
                current = i
                break

        new = current + 1
        if new > len(self.sources) - 1:
            new = 0

        self.proxy.ActivateInputSourceIndex("(i)", self.sources[new].index)

    def exit_expo_and_overview(self):
        if self.ensure_proxy_alive():
            self.proxy.set_property("overview-active", False)
            self.proxy.set_property("expo-active", False)

    def on_failure(self, *args):
        print("Failed to connect to Cinnamon - screensaver will not activate when expo or overview modes are active.", flush=True)



#!/usr/bin/python3

import gi
gi.require_version('Gtk', '3.0')
gi.require_version('GdkX11', '3.0')
gi.require_version('CScreensaver', '1.0')

from gi.repository import Gtk, Gdk, CScreensaver, Gio

import signal
import gettext
import argparse
import os
import setproctitle
import sys

import config
import status
from util import utils, settings
from service import ScreensaverService

signal.signal(signal.SIGINT, signal.SIG_DFL)
gettext.install("cinnamon-screensaver", "/usr/share/locale")

class Main(Gtk.Application):
    """
    This is the main entry point to the program, and it shows up
    in the process list.  We do any theme preparation here as well.

    We start the ScreensaverService from here.
    """
    def __init__(self):
        super(Main, self).__init__(application_id="org.cinnamon.ScreenSaver",
                                   inactivity_timeout=30000,
                                   flags=Gio.ApplicationFlags.IS_SERVICE)

        # Our service must be set up before we register with the session manager.
        ScreensaverService()

    def do_activate(self):
        pass

    def do_startup(self):
        print("Starting screensaver...", flush=True)
        Gtk.Application.do_startup(self)

        parser = argparse.ArgumentParser(description='Cinnamon Screensaver')
        parser.add_argument('--debug', dest='debug', action='store_true',
                            help='Print out some extra debugging info')
        parser.add_argument('--interactive-debug', dest='interactive', action='store_true',
                            help='If multiple monitors are in use, only cover one monitor, and launch GtkInspector')
        parser.add_argument('--disable-locking', dest='lock_disabled', action='store_true',
                            help='Disable the lock screen')
        parser.add_argument('--version', dest='version', action='store_true',
                            help='Display the current version')
        parser.add_argument('--hold', dest='hold', action='store_true',
                            help="Keep the process running." \
                                 "Normally cinnamon-screensaver will exit after being idle for 30 seconds.")
        parser.add_argument('--no-fallback', dest='no_fallback', action='store_true',
                            help="Don't spawn a fallback window when locking the screen.")
        args = parser.parse_args()

        if settings.get_custom_screensaver() != '':
            print("custom screensaver selected, exiting cinnamon-screensaver.", flush=True)
            quit()

        if args.version:
            print("cinnamon-screensaver %s" % config.VERSION)
            quit()

        status.LockEnabled = not args.lock_disabled
        status.Debug = args.debug
        status.InteractiveDebug = args.interactive
        status.UseFallback = not args.no_fallback
        # The inactivity-timeout will be ignored until there's been an initial hold. Simply
        # starting the app and letting it idle will end up with it exiting after 10s no matter
        # what the timeout.
        self.hold()

        if not args.hold:
            self.release()

        if status.Debug:
            print("Debug mode active", flush=True)

        if args.lock_disabled:
            print("Locking disabled", flush=True)

        # This is here mainly to allow the notification watcher to have a valid status.Debug value
        import singletons
        Gtk.Settings.get_default().connect("notify::gtk-theme-name", self.on_theme_changed)
        self.do_style_overrides()

    def on_theme_changed(self, settings, pspec, data=None):
        self.do_style_overrides()

    def do_style_overrides(self):
        """
        Here we try to check for theme support in the current system's gtk theme.

        We do this by retrieving a string of the current theme's final style sheet,
        then searching for the .csstage style class.  If it's found, we return, otherwise
        we add our own application-priority provider as a fallback.  While we have the
        theme string, we check for a variable name we can use for the fallback experience,
        and adjust it in our application stylesheet if necessary before adding it as a
        provider.
        """
        theme_name = Gtk.Settings.get_default().get_property("gtk-theme-name")
        provider = Gtk.CssProvider.get_named(theme_name)

        css = provider.to_string()

        if ".csstage" not in css:
            print("Cinnamon Screensaver support not found in current theme - adding some...", flush=True)

            path = os.path.join(config.pkgdatadir, "cinnamon-screensaver.css")

            f = open(path, 'r')
            fallback_css = f.read()
            f.close()

            if "@define-color theme_selected_bg_color" in css:
                pass
            elif "@define-color selected_bg_color" in css:
                print("replacing theme_selected_bg_color with selected_bg_color", flush=True)
                fallback_css = fallback_css.replace("@theme_selected_bg_color", "@selected_bg_color")
            else:
                print("replacing theme_selected_bg_color with Adwaita blue", flush=True)
                fallback_css = fallback_css.replace("@selected_bg_color", "#4a90d9")

            fallback_prov = Gtk.CssProvider()

            try:
                fallback_prov.load_from_data(fallback_css.encode())
                Gtk.StyleContext.add_provider_for_screen (Gdk.Screen.get_default(), fallback_prov, 600)
                Gtk.StyleContext.reset_widgets(Gdk.Screen.get_default())
            except Exception as e:
                print("Could not parse fallback css: %s" % str(e))

if __name__ == "__main__":
    setproctitle.setproctitle('cinnamon-screensaver')

    main = Main()
    main.run()



#!/usr/bin/python3

from gi.repository import Gio, GObject, GLib

import status
from util.utils import DEBUG

class NameBlocker(GObject.GObject):
    """
    Terminates other screensavers.  Ideally we'd just take ownership of their
    dbus names, but due to how DE detection works in xdg-screensaver, that script
    incorrectly assumes we're a gnome or mate session if their names are owned.
    """
    def __init__(self):
        super(NameBlocker, self).__init__()

        self.owned_names = []

        self.watch("org.gnome.ScreenSaver")
        self.watch("org.mate.ScreenSaver")

    def watch(self, name):
        handle = Gio.bus_watch_name(Gio.BusType.SESSION,
                                    name,
                                    Gio.BusNameWatcherFlags.NONE,
                                    self.on_name_appeared,
                                    self.on_name_lost)

        self.owned_names.append((handle, "name"))

    def unwatch_all(self):
        for (handle, name) in self.owned_names:
            DEBUG("Releasing dbus name: %s" % name)

            Gio.bus_unown_name(handle)

        self.owned_names = []

    def on_name_appeared(self, connection, name, name_owner, data=None):
        DEBUG("%s appeared on the session bus, killing it" % name)

        connection.call(name_owner,
                        "/" + name.replace(".", "/"),
                        name,
                        "Quit",
                        None,
                        None,
                        Gio.DBusCallFlags.NONE,
                        -1,
                        None)

    def on_name_lost(self, connection, name, data=None):
        DEBUG("%s is gone from the session bus" % name)

    def do_dispose(self):
        DEBUG("nameBlocker do_dispose")

        self.unwatch_all()

        super(NameBlocker, self).do_dispose()



#!/usr/bin/python3

from gi.repository import Gdk

import status
from util.keybindings import KeyBindings

MOTION_THRESHOLD = 100

class EventHandler:
    """
    The EventHandler receives all user key, button and motion events
    for the application.  At various points it can be receiving them
    from the Stage window, an offscreen window, or the root GdkWindow.

    All events are stopped from passing beyond this point, to prevent
    any unintended events from propagating to Cinnamon or Muffin.
    """
    def __init__(self, manager):
        self.manager = manager
        self.keybindings_handler = KeyBindings(manager)

        self.last_x = -1
        self.last_y = -1


    def on_user_activity(self):
        """
        Any user event is a 'wake' event, and is propagated to the stage
        in order to reset our unlock cancellation timer.
        """
        self.manager.simulate_user_activity()

    def on_motion_event(self, event):
        """
        Any mouse movement is sent here - there is a threshold to reach when
        asleep, so that inadvertent motion doesn't wake the system unintentionally.
        """
        if status.Awake:
            self.on_user_activity()
            return Gdk.EVENT_PROPAGATE

        if self.last_x == -1 or self.last_y == -1:
            self.last_x = event.x
            self.last_y = event.y
            return Gdk.EVENT_PROPAGATE

        distance = max(abs(self.last_x - event.x), abs(self.last_y - event.y))

        if distance > MOTION_THRESHOLD:
            self.on_user_activity()

        return Gdk.EVENT_PROPAGATE

    def on_button_press_event(self, event):
        """
        Any button presses are swallowed after interacting
        with their receiving widgets (in the case of buttons, entry, etc...)
        """
        self.on_user_activity()

        return Gdk.EVENT_STOP

    def on_key_press_event(self, event):
        """
        Any key events are checked with the KeybindingHandler in case
        a media shortcut is being used, or it's a special keystroke such
        as Escape or tab.

        Any other presses are sent to the password entry or swallowed.
        """
        if self.keybindings_handler.maybe_handle_event(event):
            return Gdk.EVENT_STOP

        if status.Active:
            if status.Locked:
                self.manager.queue_dialog_key_event(event)

        self.on_user_activity()

        return Gdk.EVENT_STOP



#!/usr/bin/python3

from gi.repository import Gdk, GObject, GLib, Gio, CScreensaver
import time
import traceback
import os
import signal
import subprocess

import config
import constants as c
import status

from stage import Stage
import singletons
from util import utils, settings, trackers
from util.focusNavigator import FocusNavigator
from util.utils import DEBUG

class ScreensaverManager(GObject.Object):
    """
    The ScreensaverManager is the central point where most major decision are made,
    and where ScreensaverService requests are acted upon.
    """
    __gsignals__ = {
        'active-changed': (GObject.SignalFlags.RUN_LAST, None, (bool, )),
    }

    def __init__(self):
        super(ScreensaverManager, self).__init__()

        self.activated_timestamp = 0

        self.stage = None
        self.old_stage = None
        self.stage_refresh_id = 0
        self.refreshing = False
        self.refresh_again = False

        self.fb_pid = 0
        self.fb_failed_to_start = False

        # Ensure our state
        status.Active = False
        status.Locked = False
        status.Awake = False

        self.focus_nav = FocusNavigator()

        self.session_client = singletons.SessionClient
        trackers.con_tracker_get().connect(self.session_client,
                                           "idle-changed",
                                           self.on_session_idle_changed)

        self.cinnamon_client = singletons.CinnamonClient

        singletons.LoginClientResolver(self)

    def is_locked(self):
        """
        Return if we're Locked - we could be Active without being locked.
        """
        return status.Locked

    def set_locked(self, locked):
        if locked:
            status.Locked = True
            if status.UseFallback:
                self.spawn_fallback_window()
        else:
            status.Locked = False
            if status.UseFallback:
                self.kill_fallback_window()

    def lock(self, msg=""):
        """
        Initiate locking (activating first if necessary.)  Return True if we were
        already active and just need to set the lock flag (or we were already locked
        as well).  Return False if we're not active, and need to construct a stage, etc...
        """
        if not status.Active:
            if self.set_active(True, msg):
                self.stop_lock_delay()
                if utils.user_can_lock():
                    self.set_locked(True)
                return False
        else:
            if utils.user_can_lock():
                self.set_locked(True)
            self.stage.set_message(msg)

        # Return True to complete any invocation immediately because:
        # - we were already active and possibly already locked
        # - we were unable to achieve a server grab, or something else
        #   prevents the activation from proceeding, and we don't want to
        #   block anything...??
        return True

    def unlock(self):
        """
        Initiate unlocking and deactivating
        """
        self.set_active(False)
        self.set_locked(False)
        status.Awake = False

    def set_active(self, active, msg=None):
        """
        Activates or deactivates the screensaver.  Activation involves:
            - sending a request to Cinnamon to exit Overview or Expo -
              this could prevent a successful screen grab and keep the
              screensaver from activating.
            - grabbing the keyboard and mouse.
            - creating the screensaver Stage.
        Deactivation involves:
            - destroying the screensaver stage.
            - releasing our keyboard and mouse grabs.
        """
        if active:
            if not status.Active:
                self.cinnamon_client.exit_expo_and_overview()
                self.grab_helper = CScreensaver.EventGrabber.new(status.Debug)
                status.screen = CScreensaver.Screen.new(status.Debug)

                if self.grab_helper.grab_root(False):
                    if not self.stage:
                        Gio.Application.get_default().hold()
                        self.spawn_stage(msg, self.on_spawn_stage_complete)
                    else:
                        self.stage.activate(self.on_spawn_stage_complete)
                        self.stage.set_message(msg)
                    return True
                else:
                    status.Active = False
                    return False
            else:
                self.stage.set_message(msg)
                return True
        else:
            if self.stage:
                self.grab_helper.release()
                self.despawn_stage()
                Gio.Application.get_default().release()
                status.focusWidgets = []
            status.screen = None
            return True

    def get_active(self):
        """
        Return whether we're Active or not (showing) - this is not
        necessarily Locked.
        """
        return status.Active

    def get_active_time(self):
        """
        Return how long we've been activated, or 0 if we're not
        """
        if self.activated_timestamp != 0:
            return int(time.time() - self.activated_timestamp)
        else:
            return 0

    def simulate_user_activity(self):
        """
        Called upon any key, motion or button event, does different things
        depending on our current state.

        If we're idle:
            - do nothing

        If we're locked:
            - show the unlock widget (if it's already visible, this also has
              the effect of resetting the unlock timeout - see Stage.py)
            - show the mouse pointer, so the user can navigate the unlock screen.

        If we're Active but not Locked, simply deactivate (destroying the Stage
        and returning the screensaver back to idle mode.)
        """
        if not status.Active:
            return

        if not status.Awake:
            DEBUG("*******************************************")
            DEBUG("manager: user activity, waking:")
            # traceback.print_stack()
            DEBUG("*******************************************")

        if status.Locked and self.stage.initialize_pam():
            if not status.Awake:
                DEBUG("manager: locked, raising unlock widget")

            self.stage.raise_unlock_widget()
            self.grab_helper.release_mouse()
            self.stage.maybe_update_layout()
        else:
            DEBUG("manager: not locked, queueing idle deactivation")

            trackers.timer_tracker_get().add_idle("idle-deactivate",
                                                  self.idle_deactivate)

    def idle_deactivate(self):
        self.set_active(False)

        trackers.timer_tracker_get().cancel("idle-deactivate")

        return False

    def spawn_stage(self, away_message, callback=None):
        """
        Create the Stage and begin fading it in.  This may run quickly, in the case of
        user-initiated activation, or slowly, when the session has gone idle.
        """
        try:
            self.stage = Stage(self, away_message)
            self.stage.activate(callback)
        except Exception:
            print("Could not spawn screensaver stage:\n")
            traceback.print_exc()
            self.grab_helper.release()
            status.Active = False
            self.cancel_timers()

    def spawn_fallback_window(self):
        if self.fb_pid > 0:
            return

        DEBUG("manager: spawning fallback window")

        if self.stage.get_realized():
            self._real_spawn_fallback_window(self)
        else:
            self.stage.connect("realize", self._real_spawn_fallback_window)

    def get_tty_vals(self):
        session_tty = None
        term_tty = None
        username = GLib.get_user_name()[:8]
        used_tty = []

        try:
            tty_output = subprocess.check_output(["w", "-h"]).decode("utf-8")
            for line in tty_output.split("\n"):
                if line.startswith(username):
                    if "cinnamon-session" in line and "tty" in line:
                        session_tty = line.split()[1].replace("tty", "")
                        used_tty.append(session_tty)
                    elif "tty" in line:
                        term_tty = line.split()[1].replace("tty", "")
                elif "tty" in line:
                    used_tty.append(line.split()[1].replace("tty", ""))

            used_tty.sort()

            if term_tty is None:
                for i in range(1, 6):
                    if str(i) not in used_tty:
                        term_tty = str(i)
                        break
        except Exception as e:
            print("Failed to get tty numbers using w -h: %s" % str(e))

        if session_tty is None:
            try:
                session_tty = os.environ["XDG_VTNR"]
            except KeyError:
                session_tty = "7"

        if term_tty is None:
            term_tty = "2" if session_tty != "2" else "1"

        return [term_tty, session_tty]

    def _real_spawn_fallback_window(self, stage, data=None):
        if self.fb_pid > 0:
            return

        term_tty, session_tty = self.get_tty_vals()

        argv = [
            os.path.join(config.pkglibdir, "cs-backup-locker"),
            "--xid", str(self.stage.get_window().get_xid()),
            "--term", term_tty,
            "--session", session_tty
        ]

        if status.Debug:
            argv.append("--debug")

        try:
            self.fb_pid = GLib.spawn_async(argv)[0]
        except GLib.Error as e:
            self.fb_failed_to_start = True
            print("Could not start screensaver fallback process: %s" % e.message, flush=True)

        try:
            self.stage.disconnect_by_func(self._real_spawn_fallback_window)
        except:
            pass

    def kill_fallback_window(self):
        if self.fb_pid == 0 and not self.fb_failed_to_start:
            return

        DEBUG("manager: killing fallback window")

        try:
            DEBUG("manager: checking if fallback window exists first.")
            if self.fb_pid > 0:
                os.kill(self.fb_pid, 0)
            elif self.fb_failed_to_start:
                raise ProcessLookupError("Fallback window failed to start")
        except ProcessLookupError:
            DEBUG("manager: fallback window terminated before the main screensaver, something went wrong!")
            notification = Gio.Notification.new(_("Cinnamon Screensaver has experienced an error"))

            notification.set_body(_("The 'cs-backup-locker' process terminated before the screensaver did. "
                                    "Please report this issue and try to describe any actions you may "
                                    "have performed prior to this occurring."))
            notification.set_icon(Gio.ThemedIcon(name="dialog-error"))
            notification.set_priority(Gio.NotificationPriority.URGENT)
            Gio.Application.get_default().send_notification("cinnamon-screensaver", notification)

        try:
            os.kill(self.fb_pid, signal.SIGTERM)
        except:
            pass

        self.fb_failed_to_start = False
        self.fb_pid = 0

    def on_spawn_stage_complete(self):
        """
        Called after the stage become visible.  All user events are now
        redirected to GrabHelper, our status is updated, our active timer
        is started, and emit an active-changed signal (Which is listened to
        by our ConsoleKit client if we're using it, and our own ScreensaverService.)
        """
        self.grab_stage()
        self.stage.connect("needs-refresh", self.queue_refresh_stage)
        if self.old_stage:
            self.old_stage.hide()
            self.old_stage.destroy_stage()
            self.old_stage = None
            self.stage_refreshed()

        status.Active = True

        self.emit("active-changed", True)

        self.start_timers()

    def despawn_stage(self):
        """
        The stage is destroyed, our status
        is updated, timer is canceled and active-changed is fired.
        """
        self.stage.cancel_unlocking()
        self.stage.hide()

        was_active = status.Active == True
        status.Active = False

        if was_active:
            self.emit("active-changed", False)

        self.cancel_timers()

        self.stage.destroy_stage()
        self.stage = None

        # Ideal time to check for leaking connections that might prevent GC by python and gobject
        if status.Debug:
            trackers.con_tracker_get().dump_connections_list()
            trackers.timer_tracker_get().dump_timer_list()

    def grab_stage(self):
        """
        Makes a hard grab on the Stage window, all keyboard and mouse events are dispatched or eaten
        by us now.
        """
        if self.stage is not None:
            self.grab_helper.move_to_window(self.stage.get_window(), Gdk.Screen.get_default(),  True)

    def queue_refresh_stage(self, stage):
        """
        Queues a complete refresh of the stage, resizing the screen if necessary,
        reconstructing the individual monitor objects, etc...
        """
        if self.stage_refresh_id > 0:
            GObject.source_remove(self.stage_refresh_id)
            self.stage_refresh_id = 0

        DEBUG("manager: queuing stage refresh")

        # self.stage_refresh_id = GLib.timeout_add_seconds(3, self._update_full_stage_on_idle, priority=GLib.PRIORITY_DEFAULT)
        self.stage_refresh_id = GLib.idle_add(self._update_full_stage_on_idle, priority=GLib.PRIORITY_DEFAULT)

    def _update_full_stage_on_idle(self, data=None):
        self.stage_refresh_id = 0

        if self.refreshing:
            self.refresh_again = True
            return

        Gdk.flush()

        self.refreshing = True
        self.refresh_stage()

        return False

    def stage_refreshed(self):
        DEBUG("manager: stage refresh complete")

        self.grab_stage()

        self.refreshing = False
        if self.refresh_again:
            self.refresh_again = False

            DEBUG("Got refresh signal while refreshing, doing it again.")
            self.queue_refresh_stage()

        self.simulate_user_activity()

    def refresh_stage(self, stage=None):
        """
        Tells the stage to check its canvas size and make sure its windows are up-to-date.  This is called
        when our login manager tells us its "Active" property has changed.  We are always connected to the
        login manager, so we first check if we have a stage.
        """
        if self.stage is None:
            return

        DEBUG("Stage: refreshing")

        self.stage.cancel_unlocking()

        self.cancel_timers()
        status.focusWidgets = []
        self.grab_helper.release()
        away_message = self.stage.away_message

        self.old_stage = self.stage
        self.spawn_stage(away_message, self.on_spawn_stage_complete)

    def start_timers(self):
        """
        Stamps our current time starts our lock delay timer (the elapsed time to allow after
        activation, to lock the computer.)
        """
        self.activated_timestamp = time.time()
        self.start_lock_delay()

    def cancel_timers(self):
        """
        Zeros out our activated timestamp and cancels our lock delay timer.
        """
        self.activated_timestamp = 0
        self.stop_lock_delay()

    def cancel_unlocking(self):
        """
        Return to sleep (not Awake) - hides the pointer and the unlock widget.
        """
        self.grab_stage()
        self.stage.cancel_unlocking()

    def on_lock_delay_timeout(self):
        """
        Updates the lock status when our timer has hit its limit
        """
        DEBUG("manager: locking after delay ('lock-delay')")

        self.set_locked(True)

        return False

    def start_lock_delay(self):
        """
        Setup the lock delay timer based on user prefs - if there is
        no delay, or if idle locking isn't enabled, we run the callback
        immediately, or simply return, respectively.
        """
        if not settings.get_idle_lock_enabled():
            return

        if not utils.user_can_lock():
            return

        lock_delay = settings.get_idle_lock_delay()

        if lock_delay == 0:
            self.on_lock_delay_timeout()
        else:
            trackers.timer_tracker_get().start_seconds("idle-lock-delay",
                                                       lock_delay,
                                                       self.on_lock_delay_timeout)

    def stop_lock_delay(self):
        """
        Cancels the lock delay timer.
        """
        trackers.timer_tracker_get().cancel("idle-lock-delay")

##### EventHandler/GrabHelper/FocusNavigator calls.

    def queue_dialog_key_event(self, event):
        """
        Forwards a captured key event to the stage->unlock dialog.
        """
        self.stage.queue_dialog_key_event(event)

    def propagate_tab_event(self, shifted):
        """
        Forwards a tab event to the focus navigator.
        """
        self.focus_nav.navigate(shifted)

    def propagate_activation(self):
        """
        Forwards an activation event (return) to the focus navigator.
        """
        self.focus_nav.activate_focus()

    def get_focused_widget(self):
        """
        Returns the currently focused widget from the FocusNavigator
        """
        return self.focus_nav.get_focused_widget()

    def on_session_idle_changed(self, proxy, idle):
        """
        Call back for the session client - initiates a slow fade-in
        for the stage when the session goes idle.  Cancels the stage fade-in
        if idle becomes False before it has completed its animation.
        """
        DEBUG("manager: session idle-changed - %s" % str(idle))

        if idle and not status.Active:
            self.set_active(True)



#!/usr/bin/python3

import gi
gi.require_version('CDesktopEnums', '3.0')
from gi.repository import Gtk, GObject, Gdk, Gio, CinnamonDesktop
from gi.repository.CDesktopEnums import MediaKeyType as MK

import status
import singletons
from util import settings

ALLOWED_ACTIONS = [MK.MUTE,
                   MK.VOLUME_UP,
                   MK.VOLUME_UP_QUIET,
                   MK.VOLUME_DOWN,
                   MK.VOLUME_DOWN_QUIET,
                   MK.MIC_MUTE,
                   MK.EJECT,
                   MK.PLAY,
                   MK.PAUSE,
                   MK.STOP,
                   MK.PREVIOUS,
                   MK.NEXT,
                   MK.REWIND,
                   MK.FORWARD,
                   MK.REPEAT,
                   MK.RANDOM,
                   MK.TOUCHPAD,
                   MK.TOUCHPAD_ON,
                   MK.TOUCHPAD_OFF,
                   MK.SHUTDOWN,
                   MK.SUSPEND,
                   MK.HIBERNATE,
                   MK.SCREEN_BRIGHTNESS_UP,
                   MK.SCREEN_BRIGHTNESS_DOWN,
                   MK.ROTATE_VIDEO,
                   MK.KEYBOARD_BRIGHTNESS_UP,
                   MK.KEYBOARD_BRIGHTNESS_DOWN,
                   MK.KEYBOARD_BRIGHTNESS_TOGGLE]

class ShortcutAction(GObject.GObject):
    """
    A class representing a group of keybindings for a specific
    media key action.
    """
    def __init__(self, action, bindings):
        super(ShortcutAction, self).__init__()

        self.action = action
        self.bindings = bindings

        self.parsed = []

        for binding in self.bindings:
            key, codes, mods = Gtk.accelerator_parse_with_keycode(binding)

            self.parsed.append((key, codes, mods))

    def activate(self, key, keycode, mods):
        for binding in self.parsed:
            if (key == binding[0] or keycode in binding[1]) and mods == binding[2]:
                return self.action

        return -1

class KeyBindings(GObject.GObject):
    """
    Receives keystrokes from the EventHandler - it checks for media key
    activation, as well as special keys like Escape or Tab, and performs the
    appropriate actions.
    """
    def __init__(self, manager):
        super(KeyBindings, self).__init__()

        self.manager = manager

        self.client = singletons.KeybindingHandlerClient

        self.keymap = Gdk.Keymap.get_default()

        self.media_key_settings = Gio.Settings(schema_id="org.cinnamon.desktop.keybindings.media-keys")
        self.shortcut_actions = []

        self.load_bindings()

    def load_bindings(self):
        self.shortcut_actions = []

        for action_id in ALLOWED_ACTIONS:
            bindings = self.media_key_settings.get_strv(CinnamonDesktop.desktop_get_media_key_string(action_id))

            action = ShortcutAction(action_id, bindings)

            self.shortcut_actions.append(action)

    def maybe_handle_event(self, event):
        if event.type != Gdk.EventType.KEY_PRESS:
            return False

        filtered_state = Gdk.ModifierType(event.state & ~(Gdk.ModifierType.MOD2_MASK | Gdk.ModifierType.LOCK_MASK))

        if filtered_state == 0 and event.keyval == Gdk.KEY_Escape:
            if status.Awake:
                self.manager.cancel_unlocking()
                return True

        if event.keyval in (Gdk.KEY_Menu, Gdk.KEY_F10):
            return True

        if status.Awake:
            if event.keyval in (Gdk.KEY_Tab, Gdk.KEY_ISO_Left_Tab):
                if event.keyval == Gdk.KEY_ISO_Left_Tab:
                    self.manager.propagate_tab_event(True)
                else:
                    self.manager.propagate_tab_event(False)
                return True
            elif (event.keyval in (Gdk.KEY_space, Gdk.KEY_Return, Gdk.KEY_KP_Enter)) and \
                  isinstance(self.manager.get_focused_widget(), Gtk.Button):
                self.manager.propagate_activation()
                return True

        if settings.get_allow_shortcuts():
            for entry in self.shortcut_actions:
                res = entry.activate(event.keyval, event.hardware_keycode, filtered_state)

                if res == -1:
                    continue
                else:
                    self.client.handle_keybinding(res)
                    return True

        return False



#!/usr/bin/python3

import gi
gi.require_version('CScreensaver', '1.0')

from gi.repository import GLib, CScreensaver, Gio
import os
import sys
import signal
import argparse
import gettext
import shlex
from enum import IntEnum
from subprocess import Popen, DEVNULL

import config
from util import settings
import constants as c

signal.signal(signal.SIGINT, signal.SIG_DFL)
gettext.install("cinnamon-screensaver", "/usr/share/locale")

class Action(IntEnum):
    EXIT = 1
    QUERY = 2
    TIME = 3
    LOCK = 4
    ACTIVATE = 5
    DEACTIVATE = 6
    VERSION = 7

class ScreensaverCommand:
    """
    This is a standalone executable that provides a simple way
    of controlling the screensaver via its dbus interface.
    """
    def __init__(self, mainloop):
        self.mainloop = mainloop
        self.proxy = None

        parser = argparse.ArgumentParser(description='Cinnamon Screensaver Command')
        parser.add_argument('--exit', '-e', dest="action_id", action='store_const', const=Action.EXIT,
                            help=_('Causes the screensaver to exit gracefully'))
        parser.add_argument('--query', '-q', dest="action_id", action='store_const', const=Action.QUERY,
                            help=_('Query the state of the screensaver'))
        parser.add_argument('--time', '-t', dest="action_id", action='store_const', const=Action.TIME,
                            help=_('Query the length of time the screensaver has been active'))
        parser.add_argument('--lock', '-l', dest="action_id", action='store_const', const=Action.LOCK,
                            help=_('Tells the running screensaver process to lock the screen immediately'))
        parser.add_argument('--activate', '-a', dest="action_id", action='store_const', const=Action.ACTIVATE,
                            help=_('Turn the screensaver on (blank the screen)'))
        parser.add_argument('--deactivate', '-d', dest="action_id", action='store_const', const=Action.DEACTIVATE,
                            help=_('If the screensaver is active then deactivate it (un-blank the screen)'))
        parser.add_argument('--version', '-V', dest="action_id", action='store_const', const=Action.VERSION,
                            help=_('Version of this application'))
        parser.add_argument('--away-message', '-m', dest="message", action='store', default="",
                            help=_('Message to be displayed in lock screen'))
        args = parser.parse_args()

        if not args.action_id:
            parser.print_help()
            quit()

        if args.action_id == Action.VERSION:
            print("cinnamon-screensaver %s" % config.VERSION)
            quit()

        self.action_id = args.action_id
        self.message = args.message

        custom_saver = settings.get_custom_screensaver()
        if custom_saver != '':
            self.handle_custom_saver(custom_saver)
            quit()

        CScreensaver.ScreenSaverProxy.new_for_bus(Gio.BusType.SESSION,
                                                  Gio.DBusProxyFlags.NONE,
                                                  c.SS_SERVICE,
                                                  c.SS_PATH,
                                                  None,
                                                  self._on_proxy_ready)

    def handle_custom_saver(self, custom_saver):
        if self.action_id in [Action.LOCK, Action.ACTIVATE]:
            try:
                Popen(shlex.split(custom_saver), stdin=DEVNULL)
            except OSError as e:
                print("Error %d running %s: %s" % (e.errno, custom_saver,
                                                   e.strerror))
        else:
            print("Action not supported with custom screensaver.")

    def _on_proxy_ready(self, object, result, data=None):
        try:
            self.proxy = CScreensaver.ScreenSaverProxy.new_for_bus_finish(result)
            self.perform_action()
        except GLib.Error as e:
            print("Can't connect to screensaver: %d - %s" % (e.code, e.message))
            self.mainloop.quit()

    def perform_action(self):
        if self.action_id == Action.EXIT:
            self.proxy.call_quit_sync()
        elif self.action_id == Action.QUERY:
            if self.proxy.call_get_active_sync():
                print(_("The screensaver is active\n"))
            else:
                print(_("The screensaver is inactive\n"))
        elif self.action_id == Action.TIME:
            time = self.proxy.call_get_active_time_sync()
            if time == 0:
                print(_("The screensaver is not currently active.\n"))
            else:
                print(gettext.ngettext ("The screensaver has been active for %d second.\n", "The screensaver has been active for %d seconds.\n", time) % time)
        elif self.action_id == Action.LOCK:
            self.proxy.call_lock_sync(self.message)
        elif self.action_id == Action.ACTIVATE:
            self.proxy.call_set_active_sync(True)
        elif self.action_id == Action.DEACTIVATE:
            self.proxy.call_set_active_sync(False)

        self.mainloop.quit()

if __name__ == "__main__":
    ml = GLib.MainLoop.new(None, True)
    main = ScreensaverCommand(ml)

    ml.run()



#!/usr/bin/python3

import gi
from gi.repository import GObject, CScreensaver, Gio, GLib
import os
import time

from util import utils, trackers

class AccountsServiceClient(GObject.Object):
    """
    Singleton for working with the AccountsService, which we use
    to retrieve the user's face image and their real name.
    """
    ACCOUNTS_SERVICE = "org.freedesktop.Accounts"
    ACCOUNTS_PATH    = "/org/freedesktop/Accounts"

    __gsignals__ = {
        'accounts-ready': (GObject.SignalFlags.RUN_LAST, None, ()),
    }

    def __init__(self):
        super(AccountsServiceClient, self).__init__()

        self.accounts = None
        self.user = None

        print("Loading AccountsService")

        CScreensaver.AccountsServiceProxy.new_for_bus(Gio.BusType.SYSTEM,
                                                      Gio.DBusProxyFlags.DO_NOT_AUTO_START,
                                                      self.ACCOUNTS_SERVICE,
                                                      self.ACCOUNTS_PATH,
                                                      None,
                                                      self.on_accounts_connected)

    def on_accounts_connected(self, source, res):
        try:
            self.accounts = CScreensaver.AccountsServiceProxy.new_for_bus_finish(res)
        except GLib.Error as e:
            print(f"Could not connect to AccountsService: {e}", flush=True)
            return

        self.accounts.call_find_user_by_name(utils.get_user_name(), None, self.got_user_proxy)

    def got_user_proxy(self, source, res):
        try:
            proxy_path = self.accounts.call_find_user_by_name_finish(res)
        except GLib.Error as e:
            print(f"Could not get AccountsService User object path: {e}", flush=True)
            return

        CScreensaver.AccountsUserProxy.new_for_bus(Gio.BusType.SYSTEM,
                                                   Gio.DBusProxyFlags.NONE,
                                                   self.ACCOUNTS_SERVICE,
                                                   proxy_path,
                                                   None,
                                                   self.on_user_loaded)

    def on_user_loaded(self, source, res):
        try:
            self.user = CScreensaver.AccountsUserProxy.new_for_bus_finish(res)
        except GLib.Error as e:
            print(f"Could not create AccountsService.User: {e}", flush=True)

        print("AccountsService ready")
        self.emit("accounts-ready")

    def get_real_name(self):
        if self.user is not None:
            return self.user.get_property("real-name")

        return None

    def get_face_path(self):
        face = os.path.join(GLib.get_home_dir(), ".face")
        if os.path.exists(face):
            return face

        if self.user is not None:
            accounts_path = self.user.get_property("icon-file")
            if os.path.exists(accounts_path):
                    return accounts_path

        return None



#!/usr/bin/python3

from gi.repository import Gtk

import status
from util import utils, trackers, settings
from baseWindow import BaseWindow
from widgets.notificationWidget import NotificationWidget
from widgets.powerWidget import PowerWidget

class InfoPanel(BaseWindow):
    """
    Upper right corner panel - contains the notification counter and any
    battery indicator(s) - this panel will generally show if it has anything
    relevant to say, regardless of our Awake state.
    """
    def __init__(self):
        super(InfoPanel, self).__init__()
        self.monitor_index = status.screen.get_primary_monitor()

        self.update_geometry()

        if not settings.get_show_info_panel():
            self.disabled = True
            return

        self.show_power = False
        self.show_notifications = False

        self.box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.box.set_halign(Gtk.Align.FILL)
        self.box.get_style_context().add_class("toppanel")
        self.box.get_style_context().add_class("infopanel")
        self.add(self.box)

        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self.box.pack_start(hbox, False, False, 6)

        self.notification_widget = NotificationWidget()
        self.notification_widget.set_no_show_all(True)
        hbox.pack_start(self.notification_widget, True, True, 2)
        self.notification_widget.connect("notification", self.on_notification_received)

        self.power_widget = PowerWidget()
        self.power_widget.set_no_show_all(True)
        self.power_widget.connect("power-state-changed",self.on_power_state_changed)
        hbox.pack_start(self.power_widget, True, True, 2)

        self.show_all()

    def refresh_power_state(self):
        if self.disabled:
            return

        self.power_widget.refresh()

    def on_notification_received(self, obj):
        self.update_visibility()

    def on_power_state_changed(self, obj):
        self.update_visibility()

    def update_visibility(self):
        """
        Determines whether or not to show the panel, depending on:
            - Whether the power widget should show (are we on battery?)
            - Whether the notification widget should show (are there any?)

        The panel will show if either of its child indicators has useful info.
        """
        if self.disabled:
            return

        do_show = False
        battery_critical = False

        self.show_power = self.power_widget.should_show()
        if self.show_power:
            battery_critical = self.power_widget.battery_critical

        self.show_notifications = self.notification_widget.should_show()

        # Determine if we want to show all the time or only when status.Awake
        if status.Awake:
            if self.show_power or self.show_notifications:
                do_show = True
        elif self.show_notifications or battery_critical:
                do_show = True

        self.set_visible(do_show)
        self.power_widget.set_visible(self.show_power)
        self.notification_widget.set_visible(self.show_notifications)
